from pathlib import Path


def find_pvi_yaml(yaml_name: str, yaml_paths: list[Path]) -> Path | None:
    """Find a yaml file in given directory"""
    for yaml_path in yaml_paths:
        if yaml_path.is_dir():
            if yaml_name in [f.name for f in yaml_path.iterdir()]:
                return yaml_path / yaml_name
    return None



"""Top level API.

.. data:: __version__
    :type: str

    Version number as calculated by https://github.com/pypa/setuptools_scm
"""

from . import device
from ._version import __version__

__all__ = ["__version__", "device"]



from pathlib import Path
from typing import Annotated, Any, Union

from pydantic import Field, TypeAdapter

from pvi._yaml_utils import YamlValidatorMixin
from pvi.device import Device, DeviceRef, enforce_pascal_case
from pvi.typed_model import TypedModel, as_tagged_union


class IndexEntry(TypedModel):
    """A structure defining an index button to launch a UI with some macros."""

    label: Annotated[
        str,
        Field(description="Button label (this will be converted to PascalCase)"),
    ]
    ui: Annotated[str, Field(description="File name of UI to open with button")]
    macros: Annotated[dict[str, str], Field(description="Macros to launch UI with")]


class Formatter(TypedModel, YamlValidatorMixin):
    """Base UI formatter."""

    @classmethod
    def type_adapter(cls) -> TypeAdapter["Formatter"]:
        """Create TypeAdapter of all child classes"""
        return TypeAdapter(
            as_tagged_union(Union[tuple(cls.__subclasses__())])  # type: ignore # noqa: UP007
        )

    @classmethod
    def from_dict(cls, serialized: dict[str, Any]) -> "Formatter":
        """Instantiate a Formatter child class from a dictionary.

        Args:
            serialized: Dictionary of class instance

        """
        return cls.type_adapter().validate_python(serialized)

    @classmethod
    def deserialize(cls, yaml: Path) -> "Formatter":
        """Instantiate a Formatter child class from YAML.

        Args:
            yaml: Path of YAML file

        """
        serialized = cls.validate_yaml(yaml)
        return cls.from_dict(serialized)

    @classmethod
    def create_schema(cls) -> dict[str, Any]:
        """Create a schema of Formatter child classes.

        Formatter itself is not included, as it should not be instanstiated directly.

        """
        cls.rebuild_child_models()
        return cls.type_adapter().json_schema()

    def format(self, device: Device, path: Path) -> None:
        """To be implemented by child classes to define how to format specific UIs.

        Args:
            device: Device to populate UI from
            path: Output file path to write UI to

        """
        raise NotImplementedError(self)

    def format_index(self, label: str, index_entries: list[IndexEntry], path: Path):
        """Format an index of buttons to open the given UIs.

        Args:
            label: Title of generated index UI
            index_entries: Buttons to format on the index UI
            path: Output path of generated UI

        """
        self.format(
            Device(
                label=label,
                children=[
                    DeviceRef(
                        name=enforce_pascal_case(index.label),
                        label=index.label,
                        pv=index.label.upper(),
                        ui=index.ui,
                        macros=index.macros,
                    )
                    for index in index_entries
                ],
            ),
            path=path,
        )



import re
from enum import Enum
from functools import cached_property
from typing import Annotated

from pydantic import BaseModel, Field

from pvi.device import ComponentUnion


class TypeStrings(BaseModel):
    """The type strings for record dtypes and parameter names"""

    asyn_read: Annotated[str, Field(description="e.g. asynInt32, asynOctetRead")]
    asyn_write: Annotated[str, Field(description="e.g. asynInt32, asynOctetWrite")]
    asyn_param: Annotated[str, Field(description="e.g. asynParamInt32, asynParamOctet")]


class Access(Enum):
    """What access does the user have. One of:

    - R: Read only value that cannot be set. E.g. chipTemperature on a detector,
      or isHomed for a motor
    - W: Write only value that can be written to, but there is no current value.
      E.g. reboot on a detector, or overwriteCurrentPosition for a motor
    - RW: Read and Write value that can be written to and read back.
      E.g. acquireTime on a detector, or velocity of a motor
    """

    R = "R"  #: Read record only
    W = "W"  #: Write record only
    RW = "RW"  #: Read and write record

    def needs_read_record(self):
        return self != self.W

    def needs_write_record(self):
        return self != self.R


class DisplayForm(Enum):
    """Instructions for how a number should be formatted for display"""

    #: Use the default representation from value
    DEFAULT = "Default"
    #: Force string representation, most useful for array of bytes
    STRING = "String"
    #: Binary, precision determines number of binary digits
    BINARY = "Binary"
    #: Decimal, precision determines number of digits after decimal point
    DECIMAL = "Decimal"
    #: Hexadecimal, precision determines number of hex digits
    HEX = "Hex"
    #: Exponential, precision determines number of digits after decimal point
    EXPONENTIAL = "Exponential"
    #: Exponential where exponent is multiple of 3, precision determines number of
    #: digits after decimal point
    ENGINEERING = "Engineering"


MACRO_RE = re.compile(r"\$\([^\)]*\)")


class Record(BaseModel):
    pv: str  # The pv of the record e.g. $(P)$(M)Status
    type: str  # The record type string e.g. ao, stringin
    fields: dict[str, str]  # The record fields
    infos: dict[str, str]  # Any infos to be added to the record

    @cached_property
    def name(self) -> str:
        """Return pv with macros removed to use as label on UIs."""
        return re.sub(MACRO_RE, "", self.pv)


class Parameter(BaseModel):
    """Base class representing an Asyn parameter"""

    invalid: list[str] = ["DESC", "DTYP", "INP", "OUT", "PINI", "VAL"]

    def _remove_invalid(self, fields: dict[str, str]) -> dict[str, str]:
        valid_fields = {
            key: value for (key, value) in fields.items() if key not in self.invalid
        }
        return valid_fields

    def generate_component(self) -> ComponentUnion:
        raise NotImplementedError(self)


class ReadParameterMixin:
    def _get_read_record(self) -> str | None:
        raise NotImplementedError(self)



import os
import shutil
from pathlib import Path

import pytest
from typer import Typer
from typer.testing import CliRunner

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


class Helper:
    @staticmethod
    def assert_output_matches(expected_path: Path, output_path: Path):
        if output_path.is_file():
            output_path = output_path.parent
            expected_path = expected_path.parent

        if os.environ.get("PVI_REGENERATE_OUTPUT", None):
            # We were asked to regenerate output, so copy output files to expected
            for output_file in output_path.iterdir():
                shutil.copy(output_file, expected_path / output_file.name)

        for output_file in output_path.iterdir():
            expected_child = expected_path / output_file.relative_to(output_path)
            assert expected_child.read_text() == output_file.read_text()

    @staticmethod
    def assert_cli_output_matches(
        app: Typer, expected_path: Path, cmd: str, output_path: Path, *paths: Path
    ):
        args = cmd.split() + [str(output_path)] + [str(p) for p in paths]
        result = CliRunner().invoke(app, args)
        if result.exception:
            raise result.exception

        Helper.assert_output_matches(expected_path, output_path)


@pytest.fixture
def helper():
    return Helper



from collections.abc import Sequence
from copy import deepcopy
from typing import Any

from lxml.etree import (
    SubElement,
    XMLParser,
    _Element,  # pyright: ignore [reportPrivateUsage]
    parse,
)

from pvi._format.utils import Bounds
from pvi._format.widget import UITemplate, WidgetFormatter
from pvi.device import (
    LED,
    BitField,
    CheckBox,
    ComboBox,
    TableRead,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
    WidgetUnion,
    WriteWidget,
)

BOB_TEXT_FORMATS = {
    TextFormat.decimal: "1",
    TextFormat.hexadecimal: "4",
    TextFormat.engineer: "3",
    TextFormat.exponential: "2",
    TextFormat.string: "6",
}


class BobTemplate(UITemplate[_Element]):
    """Extract and modify elements from a template .bob file."""

    def __init__(self, text: str):
        """Parse an XML string to an element tree object."""

        # Passing `remove_blank_text` means we can pretty print our additions
        self.tree = parse(text, parser=XMLParser(remove_blank_text=True))
        self.screen = self.search("Display")

    def set(
        self,
        template: _Element,
        bounds: Bounds | None = None,
        widget: WidgetUnion | None = None,
        properties: dict[str, Any] | None = None,
    ) -> _Element:
        properties = properties or {}

        if bounds:
            properties["x"] = bounds.x
            properties["y"] = bounds.y
            properties["width"] = bounds.w
            properties["height"] = bounds.h

            if isinstance(widget, BitField):
                properties["width"] = widget.number_of_bits * 20
                properties.pop("height")

        widget_type = template.attrib.get("type", "")

        t_copy = deepcopy(template)
        for item, value in properties.items():
            new_text = ""

            match widget_type, item, value:
                case "table", "pv_name", pv:
                    pva_prefix = "pva://"
                    if not pv.startswith(pva_prefix):
                        new_text = f"{pva_prefix}{pv}"  # Must include pva prefix
                    else:
                        new_text = str(pv)
                case "action_button", "file", file_name:
                    new_text = file_name
                    if not new_text.endswith(".bob"):
                        new_text += ".bob"  # Must include file extension
                case "action_button", "macros", dict():
                    macros: dict[str, str] = value  # type: ignore
                    if macros:
                        add_button_macros(t_copy, macros)
                case _:
                    new_text = str(value)

            if new_text:
                find_element(t_copy, item).text = new_text  # type: ignore

        # Add additional properties from widget
        match widget_type, widget:
            case ("combo", ComboBox() as combo_box):
                add_combo_box_items(t_copy, combo_box)
            case ("table", TableRead() | TableWrite() as table):
                add_table_columns(t_copy, table)
            case ("textentry", TextWrite(format=format)) | (
                "textupdate",
                TextRead(format=format),
            ) if format is not None:
                add_format(t_copy, BOB_TEXT_FORMATS[TextFormat(format)])
            case ("byte_monitor", BitField() as bit_field):
                add_byte_number_of_bits(t_copy, bit_field.number_of_bits)
            case _:
                pass

        return t_copy

    def search(self, search: str) -> _Element:
        """Locate and extract elements from the Element tree.

        Args:
            search: The unique name of the element to extract.
                Can be found in its name subelement.

        Returns:
            The extracted element.
        """

        tree_copy = deepcopy(self.tree)
        # 'name' is the unique ID for each element
        matches = [
            e
            for e in [
                element.getparent()
                for element in tree_copy.iter("name")
                if element.text == search
            ]
            if isinstance(e, _Element)
        ]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"

        # Isolate the screen properties
        if matches[0].tag == "display":
            for child in matches[0]:
                if child.tag == "widget":
                    matches[0].remove(child)

        return matches[0]

    def create_group(
        self,
        group_object: list[_Element],
        children: list[WidgetFormatter[_Element]],
        padding: Bounds | None = None,
    ) -> list[_Element]:
        """Create an xml group object from a list of child widgets

        Args:
            group_object: Templated group xml element
            children: List of child widgets within the group
            padding: Additional placement data to fit the children to the group object.
                Defaults to Bounds().

        Returns:
            An xml group with children attached as subelements
        """
        padding = padding or Bounds()

        assert len(group_object) == 1, (
            f"Size of group_object is {len(group_object)}, should be 1"
        )
        for c in children:
            group_object[0].append(c.format()[0])
        return group_object


def add_table_columns(widget_element: _Element, table: TableRead | TableWrite):
    if not table.widgets:
        # Default empty -> get options from pv
        return

    columns_element = SubElement(widget_element, "columns")
    for column, widget in enumerate(table.widgets):
        add_table_column(columns_element, f"Column {column + 1}", widget)

    add_editable(widget_element, editable=table.access_mode == "w")


def add_table_column(
    columns_element: _Element,
    name: str,
    widget: WidgetUnion,
):
    options: Sequence[str] = []
    if isinstance(widget, LED | CheckBox):
        options = ["false", "true"]
    elif isinstance(widget, ComboBox):
        options = widget.get_choices()

    column_element = SubElement(columns_element, "column")
    SubElement(column_element, "name").text = name
    add_editable(column_element, editable=isinstance(widget, WriteWidget))

    if options:
        options_element = SubElement(column_element, "options")
        for option in options:
            SubElement(options_element, "option").text = option


def add_button_macros(widget_element: _Element, macros: dict[str, str]):
    """Add action macros to the given element.

    Args:
        widget_element: Element to add action macros to
        macros: Macros to add to element

    """
    action_element = find_element(widget_element, "action")
    macros_element = SubElement(action_element, "macros")
    for macro, value in macros.items():
        SubElement(macros_element, macro).text = value


def add_combo_box_items(widget_element: _Element, combo_box: ComboBox):
    if not combo_box.get_choices():
        # Default empty -> get items from pv
        return

    items_element = SubElement(widget_element, "items")
    for item in combo_box.get_choices():
        SubElement(items_element, "item").text = item

    SubElement(widget_element, "items_from_pv").text = "false"


def add_editable(element: _Element, editable: bool):
    SubElement(element, "editable").text = "true" if editable else "false"


def add_format(element: _Element, format: str):
    if format:
        SubElement(element, "format").text = format


def add_byte_number_of_bits(element: _Element, number_of_bits: int):
    SubElement(element, "numBits").text = str(number_of_bits)


def find_element(root_element: _Element, tag: str, index: int = 0) -> _Element:
    """Iterate tree to find tag and replace text.

    Args:
        root_element: Root of tree to search
        tag: Tag to search for in tree
        index: Match to return if multiple matches are found
    """
    match list(root_element.iter(tag=tag)):
        case []:
            raise ValueError(f"No matches for '{tag}' in {root_element}")
        case matches:
            return matches[index]



from dataclasses import dataclass
from pathlib import Path

from jinja2 import Template

from pvi.device import (
    Device,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    walk,
)

PVI_TEMPLATE = Path(__file__).parent / "pvi.template.jinja"


@dataclass
class PviRecord:
    name: str
    pv: str
    access: str


def format_template(device: Device, pv_prefix: str, output: Path):
    records: list[PviRecord] = []
    for node in walk(device.children):
        match node:
            case SignalRW(read_pv=r, write_pv=w) as signal if r == w:
                records.append(PviRecord(signal.name, signal.write_pv, "rw"))
            case SignalRW() as signal:
                records.append(PviRecord(signal.name, signal.write_pv, "w"))
                records.append(PviRecord(signal.name, signal.read_pv, "r"))
            case SignalX() as signal:
                records.append(PviRecord(signal.name, signal.write_pv, "x"))
            case SignalW() as signal:
                records.append(PviRecord(signal.name, signal.write_pv, "w"))
            case SignalR() as signal:
                records.append(PviRecord(signal.name, signal.read_pv, "r"))
            case _:
                pass

    with output.open("w") as expanded, open(PVI_TEMPLATE) as template:
        template_txt = Template(template.read()).render(
            device=device.label, pv_prefix=pv_prefix, records=records
        )
        expanded.write(template_txt + "\n")



from pathlib import Path
from typing import Annotated

from pydantic import Field

from pvi._format.adl import AdlTemplate
from pvi._format.screen import (
    ScreenFormatterFactory,
    ScreenLayout,
    WidgetFormatterFactory,
)
from pvi._format.widget import (
    ActionWidgetFormatter,
    GroupFormatter,
    GroupType,
    LabelWidgetFormatter,
    PVWidgetFormatter,
    SubScreenWidgetFormatter,
    WidgetFormatter,
)
from pvi.device import Device

from .base import Formatter
from .utils import Bounds, with_title


class APSFormatter(Formatter):
    spacing: Annotated[int, Field(description="Spacing between widgets")] = 5
    title_height: Annotated[int, Field(description="Height of screen title bar")] = 25
    max_height: Annotated[int, Field(description="Max height of the screen")] = 900
    label_width: Annotated[
        int, Field(description="Width of the widget description labels")
    ] = 205
    widget_width: Annotated[int, Field(description="Width of the widgets")] = 100
    widget_height: Annotated[int, Field(description="Height of the widgets")] = 20

    def format(self, device: Device, path: Path) -> None:
        assert path.suffix == ".adl", "Can only write adl files"
        template = AdlTemplate((Path(__file__).parent / "aps.adl").read_text())
        layout = ScreenLayout(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=25,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=0,
            group_width_offset=0,
        )
        widget_formatter_factory = WidgetFormatterFactory(
            header_formatter_cls=LabelWidgetFormatter[str].from_template(
                template,
                search='"Heading"',
                property_map={"textix": "text"},
            ),
            label_formatter_cls=LabelWidgetFormatter[str].from_template(
                template,
                search='"Label"',
                property_map={"textix": "text"},
            ),
            led_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"LED"',
                sized=Bounds.square,
                property_map={"chan": "pv"},
            ),
            progress_bar_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ProgressBar"',
                property_map={"chan": "pv"},
            ),
            text_read_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"TextRead"',
                property_map={"chan": "pv"},
            ),
            check_box_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"CheckBox"',
                property_map={"chan": "pv"},
            ),
            toggle_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ToggleButton"',
                property_map={"chan": "pv"},
            ),
            combo_box_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ComboBox"',
                property_map={"chan": "pv"},
            ),
            text_write_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"TextWrite"',
                property_map={"chan": "pv"},
            ),
            action_formatter_cls=ActionWidgetFormatter[str].from_template(
                template,
                search='"SignalX"',
                property_map={"label": "label", "chan": "pv"},
            ),
            sub_screen_formatter_cls=SubScreenWidgetFormatter[str].from_template(
                template,
                search='"SubScreenFile"',
                property_map={"name": "file_name"},
            ),
        )

        label_background_formatter = WidgetFormatter[str].from_template(
            template, search="clr=2"
        )
        screen_title_formatter = LabelWidgetFormatter[str].from_template(
            template,
            search='"Title"',
            property_map={"textix": "text"},
        )
        group_title_formatter = LabelWidgetFormatter[str].from_template(
            template,
            search='"Group"',
            property_map={"textix": "text"},
        )
        group_box_formatter = WidgetFormatter[str].from_template(
            template, search='fill="outline"'
        )

        def create_group_widget_formatters(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[str]]:
            title_bounds = Bounds(
                x=bounds.x + layout.spacing,
                y=bounds.y + layout.spacing,
                w=bounds.w - 2 * layout.spacing,
                h=layout.group_label_height - layout.spacing,
            )
            return [
                group_box_formatter(bounds=bounds),
                label_background_formatter(bounds=title_bounds),
                group_title_formatter(bounds=title_bounds, text=title),
            ]

        def create_screen_widget_formatters(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[str]]:
            title_bounds = Bounds(x=0, y=0, w=bounds.w, h=layout.title_height)
            return [
                label_background_formatter(bounds=title_bounds),
                screen_title_formatter(bounds=title_bounds, text=title),
            ]

        formatter_factory: ScreenFormatterFactory[str] = ScreenFormatterFactory(
            screen_formatter_cls=GroupFormatter[str].from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(layout.spacing, layout.title_height),
                widget_formatter_hook=create_screen_widget_formatters,
            ),
            group_formatter_cls=GroupFormatter[str].from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(layout.spacing, layout.group_label_height),
                widget_formatter_hook=create_group_widget_formatters,
            ),
            widget_formatter_factory=widget_formatter_factory,
            layout=layout,
            base_file_name=path.stem,
        )
        title = f"{device.label}"

        screen_formatter, sub_screens = formatter_factory.create_screen_formatter(
            device.children, title
        )

        path.write_text("".join(screen_formatter.format()))
        for sub_screen_name, sub_screen_formatter in sub_screens:
            sub_screen_path = Path(path.parent / f"{sub_screen_name}{path.suffix}")
            sub_screen_path.write_text("".join(sub_screen_formatter.format()))



from __future__ import annotations

import json
import re
from collections.abc import Iterator, Sequence
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    ClassVar,
    Self,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)

from pvi._yaml_utils import YamlValidatorMixin, dump_yaml, type_first
from pvi.typed_model import TypedModel, as_tagged_union
from pvi.utils import find_pvi_yaml

PASCAL_CASE_REGEX = re.compile(r"(?<![A-Z])[A-Z]|[A-Z][a-z/d]|(?<=[a-z])\d")
NON_PASCAL_CHARS_RE = re.compile(r"[^A-Za-z0-9]")


def to_title_case(pascal_s: str) -> str:
    """Takes a PascalCaseFieldName and returns an Title Case Field Name

    :param pascal_s: e.g. PascalCaseFieldName

    :returns: Title Case converted name. E.g. Pascal Case Field Name

    """
    return PASCAL_CASE_REGEX.sub(lambda m: " " + m.group(), pascal_s)[1:]


def to_snake_case(pascal_s: str) -> str:
    """Takes a PascalCaseFieldName and returns a snake_case_field_name
    Args:
        pascal_s: E.g. PascalCaseFieldName
    Returns:
        snake_case converted name. E.g. pascal_case_field_name
    """
    return PASCAL_CASE_REGEX.sub(lambda m: "_" + m.group().lower(), pascal_s)[1:]


def enforce_pascal_case(s: str) -> str:
    """Enforce a pascal case string, removing any invalid characters.

    String will be returned unchanged if it is already valid PascalCase.

    Args:
        s: String to convert

    Returns: PascalCase string

    """
    if PASCAL_CASE_REGEX.fullmatch(s) is not None:
        # Already valid
        return s

    # Do minimal conversion to PascalCase:
    # - Remove any invalid characters
    # - Make first character upper case
    s = NON_PASCAL_CHARS_RE.sub(lambda _: "", s)
    return s[0].upper() + s[1:]


PascalStr = Annotated[str, Field(pattern="^([A-Z][a-z0-9]*)*$")]


class TextFormat(Enum):
    """Format to use for display of Text{Read,Write} widgets on a UI"""

    decimal = "decimal"
    hexadecimal = "hexadecimal"
    engineer = "engineer"
    exponential = "exponential"
    string = "string"


class AccessModeMixin(BaseModel):
    _access_mode: ClassVar[str]

    @property
    def access_mode(self) -> str:
        return self._access_mode


class ReadWidget(TypedModel, AccessModeMixin):
    _access_mode = "r"


class LED(ReadWidget):
    """LED display of a boolean PV"""


class BitField(ReadWidget):
    """LED and label for each bit of an int PV"""

    labels: Annotated[
        Sequence[str] | None,
        Field(description="Label for each bit"),
    ] = None
    number_of_bits: Annotated[
        int, Field(description="Number of bits to display", gt=0)
    ] = 8


class ProgressBar(ReadWidget):
    """Progress bar from lower to upper limit of a float PV"""


class TextRead(ReadWidget):
    """Text view of any PV"""

    model_config = ConfigDict(use_enum_values=True)  # Use Enum value when dumping

    lines: Annotated[int | None, Field(description="Number of lines to display")] = None
    format: Annotated[TextFormat | None, Field(description="Display format")] = None

    def get_lines(self):
        return self.lines or 1


class ArrayTrace(ReadWidget):
    """Trace of the array in a plot view"""

    axis: Annotated[
        str,
        Field(
            description=(
                "Traces with same axis name will appear on same axis, "
                "plotted against 'x' trace if it exists, or indexes if not. "
                "Only one traces with axis='x' allowed."
            ),
        ),
    ]


class ImageRead(ReadWidget):
    """2D Image view of an NTNDArray"""


class WriteWidget(TypedModel, AccessModeMixin):
    """Widget that controls a PV"""

    _access_mode = "w"


class CheckBox(WriteWidget):
    """Checkable control of a boolean PV.

    This is compact replacement for a `ToggleButton` to be used in rows and tables.

    """


class ToggleButton(WriteWidget):
    """A pair of buttons to select between two mutually exclusive states."""


class ComboBox(WriteWidget):
    """Selection of an enum PV"""

    choices: Annotated[
        Sequence[str] | None, Field(description="Choices to select from")
    ] = None

    def get_choices(self) -> list[str]:
        return [] if self.choices is None else list(self.choices)


class ButtonPanel(WriteWidget):
    """One-or-more buttons that poke a PV with a value

    Args:
        actions: Dict of button label to value the button sends

    """

    actions: Annotated[
        dict[PascalStr, str],
        Field(description="PV poker buttons"),
    ] = {"Go": "1"}


class TextWrite(WriteWidget):
    """Text control of any PV"""

    model_config = ConfigDict(use_enum_values=True)  # Use Enum value when dumping

    lines: Annotated[int | None, Field(description="Number of lines to display")] = None
    format: Annotated[TextFormat | None, Field(description="Display format")] = None

    def get_lines(self):
        return self.lines or 1


# Subset of widgets that can displayed together in one row
_RowWriteUnion = ButtonPanel | CheckBox | ComboBox | TextWrite
_RowReadUnion = BitField | LED | ProgressBar | TextRead

if not TYPE_CHECKING:
    _RowWriteUnion = as_tagged_union(_RowWriteUnion)
    _RowReadUnion = as_tagged_union(_RowReadUnion)


class ArrayWrite(WriteWidget):
    """Control of an array PV"""

    widget: Annotated[
        _RowWriteUnion, Field(description="What widget should be used for each item")
    ]


class TableRead(ReadWidget):
    """A read-only tabular view of an NTTable."""

    widgets: Annotated[
        Sequence[_RowReadUnion],
        Field(
            description="For each column, what widget should be repeated for every row",
        ),
    ] = []


class TableWrite(WriteWidget):
    """A writeable tabular view of an NTTable."""

    widgets: Annotated[
        Sequence[_RowWriteUnion],
        Field(
            description="For each column, what widget should be repeated for every row",
        ),
    ] = []


class Layout(TypedModel):
    """Widget displaying child Components"""


class Plot(Layout):
    """Children are traces of the plot"""


class Row(Layout):
    """Children are columns in the row"""

    header: Annotated[
        Sequence[str] | None, Field(description="Labels for the items in the row")
    ] = None


class Grid(Layout):
    """Children are rows in the grid"""

    labelled: Annotated[
        bool, Field(description="If True use names of children as labels")
    ] = True


class SubScreen(Layout):
    """Children are displayed on another screen opened with a button."""

    labelled: Annotated[bool, Field(description="Display labels for components")] = True


LayoutUnion = Plot | Row | Grid | SubScreen

ReadWidgetUnion = (
    ArrayTrace | BitField | ImageRead | LED | ProgressBar | TableRead | TextRead
)

WriteWidgetUnion = (
    ArrayWrite
    | ButtonPanel
    | CheckBox
    | ComboBox
    | TableWrite
    | TextWrite
    | ToggleButton
)

if not TYPE_CHECKING:
    LayoutUnion = as_tagged_union(LayoutUnion)
    ReadWidgetUnion = as_tagged_union(ReadWidgetUnion)
    WriteWidgetUnion = as_tagged_union(WriteWidgetUnion)


WidgetUnion = ReadWidgetUnion | WriteWidgetUnion


class Named(TypedModel):
    name: Annotated[
        PascalStr, Field(description="PascalCase name to uniquely identify")
    ]


class Component(Named):
    """These make up a Device"""

    label: Annotated[str | None, Field(description="Label for component")] = None
    description: Annotated[
        str | None, Field(description="Description for label tooltip")
    ] = None

    def get_label(self):
        return self.label or to_title_case(self.name)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Component) and self.name == other.name


class Signal(Component, AccessModeMixin):
    """Base signal type representing one or two PVs of a `Device`."""


class SignalR(Signal):
    """Read-only `Signal` backed by a single PV."""

    _access_mode = "r"

    read_pv: Annotated[str, Field(description="PV to use for readback")]
    read_widget: Annotated[
        ReadWidgetUnion | None,
        Field(
            description="Widget to use for display. `TextRead` will be used if unset.",
            validate_default=True,
        ),
    ] = TextRead()


class SignalW(Signal):
    """Write-only `Signal` backed by a single PV."""

    _access_mode = "w"

    write_pv: Annotated[str, Field(description="PV to use for demand")]
    write_widget: Annotated[
        WriteWidgetUnion,
        Field(description="Widget to use for control"),
    ] = TextWrite()


class SignalRW(SignalR, SignalW):
    """Read/write `Signal` backed by a demand PV and a readback PV.

    One PV can be used as both a demand and a readback by leaving `read_pv` unset. In
    this case no readback widget will be displayed unless `read_widget` is set.

    If `read_pv` is set and `read_widget` is not, a `TextRead` widget will be used.
    """

    _access_mode = "rw"
    # Flag to show this is a single PV SignalRW and the `read_pv` has been dynamically
    # set to `write_pv`. This ensures that `_validate_model` is idempotent.
    _single_pv_rw = False

    # Make optional and replace with `write_pv` during validation if not given
    read_pv: Annotated[
        str,
        Field(description="PV to use for readback. If empty, `write_pv` will be used."),
    ] = ""
    # Override description
    read_widget: Annotated[
        ReadWidgetUnion | None,
        Field(
            description=(
                "Widget to use for readback display. "
                "A `TextRead` will be used if unset and `read_pv` is given, "
                "else no readback widget will be displayed."
            ),
            validate_default=True,
        ),
    ] = None

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        if self.read_pv and not self._single_pv_rw and self.read_widget is None:
            # Update default read widget if given a read PV
            self.read_widget = TextRead()
        elif not self.read_pv:
            # No read widget, but update to use write PV for read in PVI definition
            self.read_pv = self.write_pv
            self._single_pv_rw = True

        return self


class SignalX(SignalW):
    """`SignalW` that can be triggered to write a fixed value to a PV."""

    value: Annotated[
        str,
        Field(None, description="Value to write. Uses '1' if unset."),
    ] = "1"
    write_widget: Annotated[
        WriteWidgetUnion,
        Field(description="Widget to use for actions"),
    ] = ButtonPanel()

    @model_validator(mode="after")
    def check_write_widget(self) -> Self:
        if not isinstance(self.write_widget, ButtonPanel):
            raise TypeError("SignalX write_widget must be ButtonPanel")

        return self


class DeviceRef(Component):
    """Reference to another Device."""

    pv: Annotated[str, Field(description="Child device PVI PV")]
    ui: Annotated[str, Field(description="UI file to open for referenced Device")]
    macros: Annotated[
        dict[str, str], Field(description="Macro-value pairs for UI file")
    ] = {}


class SignalRef(Component):
    """Reference to another Signal with the same name in this Device."""


class Group(Component):
    """Group of child components in a Layout"""

    layout: Annotated[
        LayoutUnion, Field(description="How to layout children on screen")
    ]
    children: Annotated[Tree, Field(description="Child Components")]


ComponentUnion = Group | SignalR | SignalW | SignalRW | SignalX | SignalRef | DeviceRef

if not TYPE_CHECKING:
    ComponentUnion = as_tagged_union(ComponentUnion)

Tree = Sequence[ComponentUnion]


def walk(tree: Tree) -> Iterator[ComponentUnion]:
    """Depth first traversal of tree"""
    for t in tree:
        if isinstance(t, Group):
            yield from walk(t.children)
        else:
            yield t


class Device(TypedModel, YamlValidatorMixin):
    """Collection of Components"""

    label: Annotated[str, Field(description="Label for screen")]
    parent: Annotated[str | None, "The parent device (basename of yaml file)"] = None
    children: Annotated[Tree, Field(description="Child Components")] = []

    def _to_dict(self) -> dict[str, Any]:
        """Serialize a `Device` instance to a `dict`.

        The returned dictionary does not include a type field for the `Device` itself
        because the discriminator is specified by the file suffix, `pvi.device.yaml`.

        In real use only `Device.serialize` will be called, which calls this method.
        This method exists as an intermediary to enable testing of the serialization
        logic separately from writing/reading a file.

        """
        d = type_first(self.model_dump(exclude_none=True))
        d.pop("type", None)
        return d

    def serialize(self, yaml: Path):
        """Serialize a `Device` instance to YAML.

        Args:
            yaml: Path of YAML file

        """
        d = self._to_dict()
        dump_yaml(d, yaml)

    @classmethod
    def deserialize(cls, yaml: Path) -> Device:
        """Instantiate a Device instance from YAML.

        Args:
            serialized: Dictionary of class instance

        """
        serialized = cls.validate_yaml(yaml)
        try:
            return cls(**serialized)
        except ValidationError:
            print(f"\nFailed to validate `{yaml}` as Device:\n")
            raise

    def deserialize_parents(self, yaml_paths: list[Path]):
        """Populate Device with Components from Device yaml of parent classes.

        Args:
            yaml_paths: Paths of parent class Device YAMLs

        """
        if self.parent is None or self.parent == "asynPortDriver":
            return

        parent_components = find_components(self.parent, yaml_paths)
        self.merge_components(parent_components)

    def merge_components(self, components: Tree) -> None:
        existing_components = list(walk(self.children))
        for node in components:
            if isinstance(node, Group):
                for group in self.children:
                    if not isinstance(group, Group):
                        continue
                    elif group.name == node.name:
                        group.children = list(group.children) + [
                            c for c in node.children if c not in group.children
                        ]
                        break  # Groups merged - skip to next group

                else:  # No break - Did not find the Group
                    # Remove any signals already in device from node
                    node.children = [
                        c for c in node.children if c not in existing_components
                    ]

                    # Add node as a new group
                    self.children = list(self.children) + [node]

                    continue  # Skip to next group
            else:
                # Node is an individual AsynParameter - just append it
                self.children = list(self.children) + [node]

    def generate_param_tree(self) -> str:
        param_tree = ", ".join(
            json.dumps(group.model_dump_json()) for group in self.children
        )
        # Encode again to quote the string as a value and escape double quotes within
        return json.dumps('{"parameters":[' + param_tree + "]}")


def find_components(yaml_name: str, yaml_paths: list[Path]) -> Tree:
    if yaml_name == "asynPortDriver":
        return []  # asynPortDriver is the most base class and has no parameters

    # Look in this module first
    device_name = f"{yaml_name}.pvi.device.yaml"
    device_yaml = find_pvi_yaml(device_name, yaml_paths)

    if device_yaml is None:
        raise OSError(f"Cannot find {device_name} in {yaml_paths}")

    device = Device.deserialize(device_yaml)

    parent_components = (
        list(find_components(device.parent, yaml_paths))
        if device.parent is not None
        else []
    )

    return list(device.children) + parent_components



import re
from typing import Annotated, Any, ClassVar, cast

from pydantic import Field

from pvi._schema_utils import rec_subclasses
from pvi.device import (
    LED,
    ComboBox,
    Named,
    ReadWidgetUnion,
    TextFormat,
    TextRead,
    TextWrite,
    ToggleButton,
    WriteWidgetUnion,
)

from ._parameters import Record, TypeStrings


class RecordError(Exception):
    pass


class AsynRecord(Record):
    def model_post_init(self, __context: Any):
        # We don't care about records without INP or OUT or with both (error)
        if all(k in self.fields.keys() for k in ("INP", "OUT")) or not any(
            k in self.fields.keys() for k in ("INP", "OUT")
        ):
            raise RecordError(f"Record {self.pv} has no input or output field or both")

        asyn_field = self.fields.get("INP", self.fields.get("OUT"))
        if asyn_field is None or "@asyn(" not in asyn_field:
            # Is it Asyn, because if it's not we don't care about it
            raise RecordError("Record has no @asyn field")

        # If there is no DESC field we must create one
        if "DESC" not in self.fields.keys():
            self.fields["DESC"] = self.name

    def get_parameter_name(self) -> str | None:
        # e.g. from: field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        # extract: FILE_PATH
        parameter_name_extractor = r"@asyn\(.*\)(\S+)"
        parameter_name = None
        for k, v in self.fields.items():
            if k == "INP" or k == "OUT":
                if "@asyn(" in v:
                    match = re.search(parameter_name_extractor, v)
                    if match:
                        parameter_name = match.group(1)
        return parameter_name

    def asyn_component_type(self) -> type["AsynParameter"]:
        # For waveform records the data type is defined by DTYP
        if self.type == "waveform":
            return get_waveform_parameter(self.fields["DTYP"])

        if "INP" in self.fields.keys():
            record_types = InRecordTypes
        elif "OUT" in self.fields.keys():
            record_types = OutRecordTypes
        else:
            raise ValueError(f"Record {self} has no 'INP' or 'OUT' field")

        try:
            return record_types[self.type]
        except KeyError as e:
            raise KeyError(
                f"{self.pv} asyn type {self.type}({self.fields}) not found in"
                f"{list(record_types)}"
            ) from e


class AsynParameter(Named):
    """Base class for all Asyn Parameters to inherit from"""

    type_strings: ClassVar[TypeStrings]
    read_record: Annotated[
        AsynRecord | None,
        Field(description="A read AsynRecord ($(name)_RBV if not given)"),
    ] = None
    write_record: Annotated[
        AsynRecord | None,
        Field(
            description="A write AsynRecord, if not given then use $(name) as write PV",
        ),
    ] = None
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()

    def get_read_pv(self) -> str:
        if self.read_record:
            return self.read_record.pv
        else:
            return self.name + "_RBV"

    def get_write_pv(self) -> str:
        if self.write_record:
            return self.write_record.pv
        else:
            return self.name


class AsynBinary(AsynParameter):
    """Asyn Binary Parameter and records"""

    type_strings = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: ReadWidgetUnion = LED()
    write_widget: WriteWidgetUnion = ToggleButton()

    def model_post_init(self, __context: Any) -> None:
        if self.write_record is not None:
            if not all(f in self.write_record.fields for f in ("ZNAM", "ONAM")):
                print(
                    f"WARNING: ZNAM/ONAM not set for {self.write_record.pv}. "
                    "Button labels will be blank."
                )


class AsynBusy(AsynBinary):
    """Asyn Busy Parameter and records"""


class AsynFloat64(AsynParameter):
    """Asyn Float64 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64",
        asyn_write="asynFloat64",
        asyn_param="asynParamFloat64",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()


class AsynInt32(AsynParameter):
    """Asyn Int32 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()


class AsynLong(AsynInt32):
    """Asyn Long Parameter and records"""


class AsynMultiBitBinary(AsynParameter):
    """Asyn MultiBitBinary Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = ComboBox()


class AsynString(AsynParameter):
    """Asyn String Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()


InRecordTypes = {
    "ai": AsynFloat64,
    "bi": AsynBinary,
    "longin": AsynLong,
    "mbbi": AsynMultiBitBinary,
    "stringin": AsynString,
}


OutRecordTypes = {
    "ao": AsynFloat64,
    "bo": AsynBinary,
    "busy": AsynBusy,
    "longout": AsynLong,
    "mbbo": AsynMultiBitBinary,
    "stringout": AsynString,
}


class AsynWaveform(AsynParameter):
    """Asyn Waveform Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    read_widget: ReadWidgetUnion = TextRead(format=TextFormat.string)
    write_widget: WriteWidgetUnion = TextWrite(format=TextFormat.string)


class AsynInt32Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with int32 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32ArrayIn",
        asyn_write="asynInt32ArrayOut",
        asyn_param="asynParamInt32",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()


class AsynFloat64Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with float64 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64ArrayIn",
        asyn_write="asynFloat64ArrayOut",
        asyn_param="asynParamFloat64",
    )
    read_widget: ReadWidgetUnion = TextRead()
    write_widget: WriteWidgetUnion = TextWrite()


WaveformRecordTypes = [AsynWaveform] + cast(
    list[type[AsynWaveform]], rec_subclasses(AsynWaveform)
)


def get_waveform_parameter(dtyp: str):
    for waveform_cls in WaveformRecordTypes:
        if dtyp in (
            waveform_cls.type_strings.asyn_read,
            waveform_cls.type_strings.asyn_write,
        ):
            return waveform_cls

    raise AssertionError(
        f"Waveform type for DTYP {dtyp} not found in {WaveformRecordTypes}"
    )



from __future__ import annotations

from collections.abc import Callable
from typing import TypeVar

from pydantic import BaseModel


class Bounds(BaseModel):
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0

    def clone(self) -> Bounds:
        return Bounds(x=self.x, y=self.y, w=self.w, h=self.h)

    def split_left(self, width: int, spacing: int) -> tuple[Bounds, Bounds]:
        """Split horizontally by width of first element"""
        to_split = width + spacing
        assert to_split < self.w, f"Can't split off {to_split} from {self.w}"
        left = Bounds(x=self.x, y=self.y, w=width, h=self.h)
        right = Bounds(x=self.x + to_split, y=self.y, w=self.w - to_split, h=self.h)
        return left, right

    def split_by_ratio(
        self, ratio: tuple[float, ...], spacing: int
    ) -> tuple[Bounds, ...]:
        """Split horizontally by ratio of widths, separated by spacing"""
        splits = len(ratio) - 1
        widget_space = self.w - splits * spacing
        widget_widths = tuple(int(widget_space * r) for r in ratio)
        widget_xs = tuple(
            self.x + sum(widget_widths[:i]) + spacing * i for i in range(splits + 1)
        )

        return tuple(
            Bounds(x=x, y=self.y, w=w, h=self.h)
            for x, w in zip(widget_xs, widget_widths, strict=True)
        )

    def split_into(self, count: int, spacing: int) -> tuple[Bounds, ...]:
        """Split horizontally into count equal widths, separated by spacing"""
        return self.split_by_ratio((1 / count,) * count, spacing)

    def square(self) -> Bounds:
        """Return the largest square that will fit in self"""
        size = min(self.w, self.h)
        return Bounds(
            x=self.x + int((self.w - size) / 2),
            y=self.y + int((self.h - size) / 2),
            w=size,
            h=size,
        )

    def added_to(self, bounds: Bounds) -> Bounds:
        return Bounds(
            x=self.x + bounds.x,
            y=self.y + bounds.y,
            w=self.w + bounds.w,
            h=self.h + bounds.h,
        )

    def tile(
        self, *, horizontal: int = 1, vertical: int = 1, spacing: int = 0
    ) -> Bounds:
        """Expand by tiling self `horizontal`/`vertical` times, plus spacing"""
        return Bounds(
            x=self.x,
            y=self.y,
            w=self.w * horizontal + spacing * (horizontal - 1),
            h=self.h * vertical + spacing * (vertical - 1),
        )

    def indent(self, indentation: int) -> None:
        self.x += indentation


T = TypeVar("T")


def concat(items: list[list[T]]) -> list[T]:
    return [x for seq in items for x in seq]


def split_with_sep(text: str, sep: str, maxsplit: int = -1) -> list[str]:
    return [t + sep for t in text.split(sep, maxsplit=maxsplit)]


def with_title(spacing: int, title_height: int) -> Callable[[Bounds], Bounds]:
    return Bounds(
        x=spacing, y=spacing + title_height, w=2 * spacing, h=2 * spacing + title_height
    ).added_to



import re
from pathlib import Path
from typing import Any, TypeVar, overload

from ruamel.yaml import YAML

T = TypeVar("T")
Leaf = dict[str, Any] | list[Any]
Branch = dict[str, "Branch | Any"] | list["Branch | Any"]
Tree = Branch | Leaf


@overload
def type_first(tree: dict[str, T]) -> dict[str, T]: ...


@overload
def type_first(tree: list[T]) -> list[T]: ...


def type_first(tree: Tree) -> Tree:
    """Walk through tree and move `type` key of dictionaries to first item.

    Args:
        node: Tree to modify

    Returns:
        The modified tree

    """
    match tree:
        case {"type": type, **rest}:
            # Move 'type' to first in dictionary
            tree = {"type": type} | rest
        case _:
            pass

    # Walk down tree
    if isinstance(tree, dict):
        for key, branch in tree.items():
            if isinstance(branch, list | dict):
                tree[key] = type_first(branch)
    else:
        for idx, branch in enumerate(tree):
            if isinstance(branch, list | dict):
                tree[idx] = type_first(branch)

    return tree


def add_line_before_type(s: str) -> str:
    return re.sub(r"([^:])(\n\s*- type:)", "\\g<1>\n\\g<2>", s)


def dump_yaml(serialized: dict[str, Any], path: Path) -> None:
    """Write serialized representation of a class instance to YAML with formatting.

    Ensure that the `type` field of every entry appears first because this
    defines what class the entry corresponds to.

    Add a space in between each entry for readability.

    """
    yaml = YAML()
    # RedHat YAML plugin settings
    yaml.indent(mapping=2, sequence=4, offset=2)  # type: ignore
    yaml.dump(serialized, path, transform=add_line_before_type)  # type: ignore


def load_yaml(path: Path) -> dict[str, Any]:
    """Load yaml from file."""
    return YAML(typ="safe").load(path)  # type: ignore


class YamlValidatorMixin:
    """Class to serialize and deserialize instances of a class to/from YAML.

    Inheriting this mixin registers that the child class can be serialized to a custom
    pvi.{cls}.yaml - where cls is the class name in lower case - and provides a utility
    to validate that a given YAML file matches the class or its child classes.

    """

    @classmethod
    def validate_yaml(cls: type, yaml: Path) -> dict[str, Any]:
        """Validate the YAML file and load into a serialized dictionary of an instance.

        This method checks that the given YAML file exists and has an appropriate file
        extension. If these checks pass the YAML file is loaded into a dictionary `d`
        and returned, such that an instance of the class could be instantiated with
        `cls(**d)`.

        Args:
            yaml: YAML representation of class to validate

        Returns: Serialized form of class instance as a dictionary

        """
        if not yaml.is_file():
            raise FileNotFoundError(yaml)

        # Walk up to find the direct child class
        while cls not in YamlValidatorMixin.__subclasses__() and len(cls.__mro__) > 1:
            cls = cls.__mro__[1]

        # Check the file extension matches the given cls
        suffix = f".pvi.{cls.__name__.lower()}.yaml"
        if not yaml.name.endswith(suffix):
            raise ValueError(f"Expected '{yaml.name}' to end with '{suffix}'")

        serialized: dict[str, Any] = load_yaml(yaml)

        cls_type: str = serialized.get("type", cls.__name__)
        if cls_type != cls.__name__:
            try:
                cls = [c for c in cls.__subclasses__() if c.__name__ == cls_type][0]
            except IndexError:
                raise ValueError(
                    f"Could not deserialize '{cls}' as subtype '{cls_type}', "
                    "not found in subclasses."
                ) from None

        return serialized



from __future__ import annotations

import re
from typing import Any

from pvi._format.utils import Bounds, split_with_sep
from pvi._format.widget import UITemplate, WidgetFormatter
from pvi.device import TextFormat, TextRead, TextWrite, WidgetUnion

ADL_TEXT_FORMATS = {
    TextFormat.decimal: "decimal",
    TextFormat.hexadecimal: "hexadecimal",
    TextFormat.engineer: "engr. notation",
    TextFormat.exponential: "exponential",
    TextFormat.string: "string",
}


class AdlTemplate(UITemplate[str]):
    def __init__(self, text: str):
        assert "children {" not in text, "Can't do groups"
        widgets = split_with_sep(text, "\n}\n")
        self.screen = "".join(widgets[:3])
        self.widgets = widgets[3:]

    def set(
        self,
        template: str,
        bounds: Bounds | None = None,
        widget: WidgetUnion | None = None,
        properties: dict[str, Any] | None = None,
    ) -> str:
        properties = properties or {}

        if bounds:
            properties["x"] = bounds.x
            properties["y"] = bounds.y
            properties["width"] = bounds.w
            properties["height"] = bounds.h

        for item, value in properties.items():
            if template.startswith('"related display"') and item == "name":
                if not value.endswith(".adl"):
                    value = f"{value}.adl"  # Must include file extension

            # Only need single line
            pattern = re.compile(rf"^(\s*{item})=.*$", re.MULTILINE)
            if isinstance(value, str):
                value = f'"{value}"'

            template, n = pattern.subn(r"\g<1>=" + str(value), template)
            assert n == 1, f"No replacements made for {item}"

        # Add additional properties from widget
        match widget:
            case TextWrite(format=format) | TextRead(format=format) if (
                is_text_widget(template) and format is not None
            ):
                template = add_property(
                    template, "format", ADL_TEXT_FORMATS[TextFormat(format)]
                )
            case _:
                pass

        return template

    def search(self, search: str) -> str:
        matches = [t for t in self.widgets if re.search(search, t)]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"
        return matches[0]

    def create_group(
        self,
        group_object: list[str],
        children: list[WidgetFormatter[str]],
        padding: Bounds | None = None,
    ) -> list[str]:
        padding = padding or Bounds()

        texts: list[str] = []

        for c in children:
            c.bounds.x += padding.x
            c.bounds.y += padding.y
            texts += c.format()

        return group_object + texts


def is_text_widget(text: str):
    return text.startswith('"text ')


def add_property(text: str, property: str, value: str) -> str:
    end = "\n}"
    return text.replace(end, f'\n\t{property}="{value}"{end}')



from pathlib import Path

import pytest
from pydantic import ValidationError

from pvi._format.base import Formatter, IndexEntry
from pvi._format.dls import DLSFormatter
from pvi._format.template import format_template
from pvi.device import (
    LED,
    ButtonPanel,
    ComboBox,
    Device,
    DeviceRef,
    Grid,
    Group,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    SubScreen,
    TableRead,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
)

HERE = Path(__file__).parent


@pytest.mark.parametrize(
    "filename,formatter",
    [
        ("text_format.adl", "aps.adl.pvi.formatter.yaml"),
        ("text_format.edl", "dls.edl.pvi.formatter.yaml"),
        ("text_format.bob", "dls.bob.pvi.formatter.yaml"),
    ],
)
def test_text_format(tmp_path, helper, filename, formatter):
    formatter_yaml = HERE / "format" / "input" / formatter
    formatter = Formatter.deserialize(formatter_yaml)

    components = []
    for format in TextFormat:
        components.append(
            SignalRW(
                name=format.name.title(),
                write_pv=f"$(P){format.name.title()}",
                write_widget=TextWrite(format=format),
                read_pv=f"$(P){format.name.title()}_RBV",
                read_widget=TextRead(format=format),
            )
        )

    device = Device(label="Text Device - $(P)", children=components)

    expected_ui = HERE / "format" / "output" / filename
    output_ui = tmp_path / filename
    formatter.format(device, output_ui)

    helper.assert_output_matches(expected_ui, output_ui)


def test_button(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    acquire_time = SignalRW(
        name="AcquireTime",
        write_pv="$(P)AcquireTime",
        write_widget=TextWrite(),
        read_pv="$(P)AcquireTime_RBV",
        read_widget=TextRead(),
    )
    acquire = SignalW(
        name="Acquire",
        write_pv="$(P)Acquire",
        write_widget=ButtonPanel(actions={"Start": "1", "Stop": "0"}),
    )
    acquire_w_rbv = SignalRW(
        name="AcquireWithRBV",
        write_pv="$(P)Acquire",
        write_widget=ButtonPanel(actions={"Start": "1", "Stop": "0"}),
        read_pv="$(P)Acquire_RBV",
        read_widget=LED(),
    )
    device = Device(
        label="Simple Device - $(P)", children=[acquire_time, acquire, acquire_w_rbv]
    )

    expected_bob = HERE / "format" / "output" / "button.bob"
    output_bob = tmp_path / "button.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_button_raises(tmp_path, helper):
    with pytest.raises(ValidationError, match="String should match pattern"):
        ButtonPanel(actions={"start": "1", "Stop": "0"})


def test_pva_table(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    table = SignalRW(
        name="PVATable",
        write_pv="$(P)PVATable",
        write_widget=TableWrite(
            widgets=[TextWrite(), ComboBox(choices=["1", "A", "True"])]
        ),
    )
    device = Device(label="TableDevice - $(P)", children=[table])

    expected_bob = HERE / "format" / "output" / "pva_table.bob"
    output_bob = tmp_path / "pva_table.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_pva_table_panda(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    table = SignalR(
        name="PandA",
        read_pv="PANDAQSRV:SEQ1:TABLE",
        read_widget=TableRead(
            widgets=[TextRead()] * 4 + [LED()] * 6 + [TextRead()] + [LED()] * 6
        ),
    )
    device = Device(label="TableDevice", children=[table])

    expected_bob = HERE / "format" / "output" / "pva_table_panda.bob"
    output_bob = tmp_path / "pva_table_panda.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_combo_box(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    combo_box = SignalRW(
        name="Enum",
        write_pv="$(P)Enum",
        write_widget=ComboBox(choices=["A", "B", "C"]),
        read_pv="$(P)Enum_RBV",
    )
    device = Device(label="Device - $(P)", children=[combo_box])

    expected_bob = HERE / "format" / "output" / "combo_box.bob"
    output_bob = tmp_path / "combo_box.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_device_ref(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    # Make a button to open an existing screen - use screen from combo box test
    device_ref = DeviceRef(
        name="ComboBox",
        pv="COMBOBOX",
        ui="combo_box",
        macros={"P": "TEST:"},
    )
    device = Device(label="Device - $(P)", children=[device_ref])

    expected_bob = HERE / "format" / "output" / "device_ref.bob"
    output_bob = tmp_path / "device_ref.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_group_sub_screen(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = Formatter.deserialize(formatter_yaml)

    # This tests every valid combination of nesting Group and SubScreen
    signals = [
        SignalR(name="A", read_pv="$(P)A", read_widget=TextRead()),
        Group(
            name="Group1",
            layout=SubScreen(),
            children=[
                SignalR(name="Group1B", read_pv="$(P)GROUP1:B", read_widget=TextRead()),
                Group(
                    name="Group2",
                    layout=Grid(),
                    children=[
                        SignalR(
                            name="Group2C",
                            read_pv="$(P)GROUP1:GROUP2:C",
                            read_widget=TextRead(),
                        ),
                        SignalR(
                            name="Group2D",
                            read_pv="$(P)GROUP1:GROUP2:D",
                            read_widget=TextRead(),
                        ),
                    ],
                ),
            ],
        ),
        Group(
            name="Group3",
            layout=Grid(),
            children=[
                SignalR(name="Group3E", read_pv="$(P)GROUP3:E", read_widget=TextRead()),
                Group(
                    name="Group4",
                    layout=SubScreen(),
                    children=[
                        SignalR(
                            name="Group4F",
                            read_pv="$(P)GROUP3:GROUP4:F",
                            read_widget=TextRead(),
                        ),
                        Group(
                            name="Group5",
                            layout=Grid(),
                            children=[
                                SignalR(
                                    name="Group5G",
                                    read_pv="$(P)GROUP3:GROUP4:GROUP5:G",
                                    read_widget=TextRead(),
                                ),
                                SignalR(
                                    name="Group5H",
                                    read_pv="$(P)GROUP3:GROUP4:GROUP5:H",
                                    read_widget=TextRead(),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        ),
    ]
    device = Device(label="Device", children=signals)

    expected_bob = HERE / "format" / "output" / "sub_screen.bob"
    output_bob = tmp_path / "sub_screen.bob"
    formatter.format(device, output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_index(tmp_path, helper):
    expected_bob = HERE / "format" / "output" / "index.bob"
    output_bob = tmp_path / "index.bob"

    DLSFormatter().format_index(
        "Index",
        [
            IndexEntry(label="Button - $(P)", ui="button.bob", macros={"P": "TEST:"}),
            IndexEntry(label="ComboBox", ui="combo_box.bob", macros={"P": "TEST:"}),
            # Check that lower case name is OK and will be capitalized to avoid
            # PascalCase validation error
            IndexEntry(label="table", ui="pva_table.bob", macros={"P": "TEST:"}),
        ],
        output_bob,
    )

    helper.assert_output_matches(expected_bob, output_bob)


def test_pvi_template(tmp_path, helper):
    read = SignalR(name="Read", read_pv="$(P)Read")
    write = SignalW(name="Write", write_pv="$(P)Write")
    read_write = SignalRW(
        name="ReadWrite", write_pv="$(P)ReadWrite", read_pv="$(P)ReadWrite_RBV"
    )
    read_write_single = SignalRW(name="ReadWriteSingle", write_pv="$(P)ReadWriteSingle")
    execute = SignalX(name="Execute", write_pv="$(P)Execute")
    device = Device(
        label="Template Device",
        children=[read, write, read_write, read_write_single, execute],
    )

    expected_bob = HERE / "format" / "output" / "pvi.template"
    output_template = tmp_path / "pvi.template"
    format_template(device, "$(P)", output_template)

    helper.assert_output_matches(expected_bob, output_template)



import re
from pathlib import Path

from pvi.device import (
    ComponentUnion,
    Grid,
    Group,
    SignalR,
    SignalRW,
    SignalW,
    Tree,
    enforce_pascal_case,
)

from ._asyn_convert import (
    AsynRecord,
    RecordError,
)
from ._parameters import Parameter

OVERRIDE_DESC = "# Overriding value in auto-generated template"


class TemplateConverter:
    def __init__(self, templates: list[Path]):
        self.templates = templates
        self._text = [t.read_text() for t in self.templates]

    def convert(self) -> Tree:
        return [
            Group(
                name=enforce_pascal_case(template.stem),
                layout=Grid(labelled=True),
                children=template_components,
            )
            for template, template_components in zip(
                self.templates, self._extract_components(), strict=True
            )
        ]

    def _extract_components(self) -> list[list[ComponentUnion]]:
        components: list[list[ComponentUnion]] = []
        for text in self._text:
            record_extractor = RecordExtractor(text)
            asyn_records = record_extractor.get_asyn_records()
            template_components: list[ComponentUnion] = []
            for parameter in RecordRoleSorter.sort_records(asyn_records):
                component = parameter.generate_component()
                template_components.append(component)
            components.append(template_components)
        return components


class RecordExtractor:
    def __init__(self, text: str):
        self._text = text

    def _extract_record_strs(self):
        # extract a whole record definition inc. fields and leading empty lines e.g.
        # record(waveform, "$(P)$(R)FilePath")
        # {
        #    field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")
        # }
        record_extractor = re.compile(r"\s*^[^#\n]*record\([^{]*{[^}]*}", re.MULTILINE)
        return re.findall(record_extractor, self._text)

    def _parse_record(self, record_str: str) -> tuple[str, str, str]:
        # extract three groups from a record definition e.g.
        # from:
        # record(waveform, "$(P)$(R)FilePath")
        # {
        #    #field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")
        # }
        # extract:
        # Group 1 - record type: waveform
        # Group 2 - record name : $(P)$(R)FilePath
        # Group 3 - all fields:
        #    #field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")

        # https://regex101.com/r/MZz1oa
        record_parser = re.compile(r'record\((\w+),\s*"?([^"]+)"?\)\s*{([^}]*)}')
        matches = re.findall(record_parser, record_str)
        if len(matches) != 1:
            raise RecordError(f"Parse failed on record: {record_str}")
        return matches[0]

    def _extract_fields(self, fields_str: str) -> list[tuple[str, str]]:
        # extract two groups from a field e.g.
        # from: field(PINI, "YES")
        # extract:
        # Group 1 - Field: PINI
        # Group 2 - Value: YES
        field_extractor = re.compile(
            r'^[^#\n]*(?:field\()([^,]*)(?:,)(?:[^"]*)(?:")([^"]*)(?:")', re.MULTILINE
        )
        return re.findall(field_extractor, fields_str)

    def _extract_infos(self, fields_str: str) -> list[tuple[str, str]]:
        # extract two groups from an info tag e.g.
        # from: info(autosaveFields, "VAL")
        # extract:
        # Group 1 - Field: autosaveFields
        # Group 2 - Value: VAL
        info_extractor = re.compile(
            r'^[^#\n]*(?:info\()([^,]*)(?:,)(?:[^"]*)(?:")([^"]*)(?:")', re.MULTILINE
        )
        return re.findall(info_extractor, fields_str)

    def _create_asyn_record(self, record_str: str) -> AsynRecord:
        record_type, record_name, record_fields = self._parse_record(record_str)

        if record_type == "motor":
            newline = "\n"
            raise RecordError(
                f"Record `{record_str.split(newline)[0]}` is type motor - ignoring"
            )

        fields = dict(self._extract_fields(record_fields))
        info = dict(self._extract_infos(record_fields))
        record = AsynRecord(pv=record_name, type=record_type, fields=fields, infos=info)
        return record

    def get_asyn_records(self) -> list[AsynRecord]:
        record_strs = self._extract_record_strs()
        record_list: list[AsynRecord] = []
        for record_str in record_strs:
            try:
                record_list.append(self._create_asyn_record(record_str))
            except RecordError as error:
                print(error)
        return record_list


class RecordRoleSorter:
    @staticmethod
    def sort_records(records: list[AsynRecord]) -> list[Parameter]:
        def _sort_inputs_outputs(
            records: list[AsynRecord],
        ) -> tuple[list[AsynRecord], list[AsynRecord]]:
            inp_records = [r for r in records if "INP" in r.fields]
            write_records = [r for r in records if "OUT" in r.fields]

            # Move waveform records with asynOctetWrite from read to write
            read_records: list[AsynRecord] = []
            for r in inp_records:
                if r.type == "waveform" and (
                    r.fields["DTYP"] == "asynOctetWrite"
                    or r.fields["DTYP"].endswith("ArrayOut")
                ):
                    write_records.append(r)
                else:
                    read_records.append(r)
            return read_records, write_records

        read_records, write_records = _sort_inputs_outputs(records)
        parameters: list[Parameter] = []
        parameters += ParameterRoleMatcher.get_actions(read_records, write_records)
        parameters += ParameterRoleMatcher.get_readbacks(read_records, write_records)
        parameters += ParameterRoleMatcher.get_setting_pairs(
            read_records, write_records
        )
        return parameters


class SettingPair(Parameter):
    read_record: AsynRecord
    write_record: AsynRecord

    def generate_component(self) -> SignalRW:
        asyn_cls = self.write_record.asyn_component_type()
        component = asyn_cls(
            name=enforce_pascal_case(self.write_record.name),
            write_record=self.write_record,
            read_record=self.read_record,
        )

        return SignalRW(
            name=component.name,
            write_pv=component.get_write_pv(),
            write_widget=component.write_widget,
            read_pv=component.get_read_pv(),
            read_widget=component.read_widget,
        )


class Readback(Parameter):
    read_record: AsynRecord

    def generate_component(self) -> SignalR:
        asyn_cls = self.read_record.asyn_component_type()

        name = self.read_record.name
        if name.endswith("_RBV"):
            name = name[: -len("_RBV")]

        component = asyn_cls(
            name=enforce_pascal_case(name), read_record=self.read_record
        )

        return SignalR(
            name=component.name,
            read_pv=component.get_read_pv(),
            read_widget=component.read_widget,
        )


class Action(Parameter):
    write_record: AsynRecord

    def generate_component(self) -> SignalW:
        asyn_cls = self.write_record.asyn_component_type()

        component = asyn_cls(
            name=enforce_pascal_case(self.write_record.name),
            write_record=self.write_record,
        )

        return SignalW(
            name=component.name,
            write_pv=component.get_write_pv(),
            write_widget=component.write_widget,
        )


class ParameterRoleMatcher:
    @staticmethod
    def get_actions(
        read_records: list[AsynRecord], write_records: list[AsynRecord]
    ) -> list[Action]:
        actions = [
            Action(write_record=w)
            for w in write_records
            if w.get_parameter_name()
            not in [r.get_parameter_name() for r in read_records]
        ]
        return actions

    @staticmethod
    def get_readbacks(
        read_records: list[AsynRecord], write_records: list[AsynRecord]
    ) -> list[Readback]:
        readbacks = [
            Readback(read_record=r)
            for r in read_records
            if r.get_parameter_name()
            not in [w.get_parameter_name() for w in write_records]
        ]
        return readbacks

    @staticmethod
    def get_setting_pairs(
        read_records: list[AsynRecord], write_records: list[AsynRecord]
    ) -> list[SettingPair]:
        setting_pairs = [
            SettingPair(read_record=r, write_record=w)
            for r in read_records
            for w in write_records
            if r.get_parameter_name() == w.get_parameter_name()
        ]
        return setting_pairs



import json
import os
from pathlib import Path
from typing import Annotated, TypeAlias

import pytest
from pydantic import BaseModel, Discriminator, Field, Tag, ValidationError

from pvi.device import (
    LED,
    CheckBox,
    ComboBox,
    Device,
    Grid,
    Group,
    SignalR,
    SignalRW,
    SignalW,
    TableWrite,
    TextRead,
    TextWrite,
)
from pvi.typed_model import TypedModel


@pytest.fixture
def device():
    components = [
        Group(
            name="Parameters",
            layout=Grid(),
            children=[
                SignalW(
                    name="WidthUnits",
                    write_pv="WIDTH:UNITS",
                    write_widget=ComboBox(),
                ),
                SignalRW(
                    name="Width",
                    write_pv="WIDTH",
                    write_widget=TextWrite(),
                    read_pv="WIDTH_RBV",
                    read_widget=TextRead(),
                ),
            ],
        ),
        SignalW(
            name="Table",
            write_pv="TABLE",
            write_widget=TableWrite(widgets=[CheckBox(), ComboBox(), TextWrite()]),
        ),
        SignalR(name="OutA", read_pv="OUTA", read_widget=LED()),
    ]
    return Device(label="label", parent="parent", children=components)


DEVICE_YAML = Path(__file__).parent / "test.pvi.device.yaml"
BAD_DEVICE_YAML = Path(__file__).parent / "bad.pvi.device.yaml"


def test_serialize(device: Device):
    if os.environ.get("PVI_REGENERATE_OUTPUT", None):
        device.serialize(DEVICE_YAML)

    s = device._to_dict()
    d = Device.validate_yaml(DEVICE_YAML)

    assert json.dumps(s, indent=2) == json.dumps(d, indent=2)


def test_deserialize(device: Device):
    d = Device.deserialize(DEVICE_YAML)
    assert d == device


def test_deserialize_raises():
    with pytest.raises(ValidationError):
        Device.deserialize(BAD_DEVICE_YAML)


def test_validate_fails():
    class NotTypedModel(BaseModel):
        pass

    BadUnion: TypeAlias = Annotated[  # type: ignore
        Annotated[SignalR, Tag("SignalR")]
        | Annotated[NotTypedModel, Tag("NotTypedModel")],
        Field(discriminator=Discriminator(TypedModel._get_type_name)),
    ]

    class Container(BaseModel):
        m: BadUnion

    # Test that trying to validate a Union using TypedModel.get_type_name fails if not
    # all members are TypedModel

    assert TypedModel._get_type_name(NotTypedModel()) is None  # type: ignore

    with pytest.raises(ValidationError):
        Container(m=NotTypedModel()).model_dump_json()



from functools import cache
from typing import TypeVar

# A type variable with an upper bound of type
Cls = TypeVar("Cls", bound=type)


# Permanently cache so we don't include deserialization subclasses defined below
@cache
def rec_subclasses(cls: Cls) -> list[Cls]:
    """Recursive implementation of type.__subclasses__"""

    subclasses: list[Cls] = []
    for sub_cls in cls.__subclasses__():
        subclasses.append(sub_cls)
        subclasses.extend(rec_subclasses(sub_cls))

    return subclasses



"""Interface for ``python -m pvi``."""

import json
import os
from pathlib import Path
from typing import Annotated, Optional

import typer

from pvi import __version__
from pvi._convert._template_convert import TemplateConverter
from pvi._convert.utils import extract_device_and_parent_class
from pvi._format import Formatter
from pvi._format.template import format_template
from pvi._pv_group import group_by_ui
from pvi.device import Device

app = typer.Typer()
convert_app = typer.Typer()
app.add_typer(convert_app, name="convert", help="Convert a module to use PVI")


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    # TODO: typer does not support `<type> | None` yet
    # https://github.com/tiangolo/typer/issues/533
    version: Optional[bool] = typer.Option(  # noqa
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    ),
):
    """PVI builder interface"""


@app.command()
def schema(
    output: Annotated[
        Path, typer.Argument(..., help="filename to write the schema to")
    ],
):
    """Write the JSON schema for the pvi interface"""
    assert output.name.endswith(".schema.json"), (
        f"Expected '{output.name}' to end with '.schema.json'"
    )

    if output.name == "pvi.device.schema.json":
        schema = Device.model_json_schema()
    elif output.name == "pvi.formatter.schema.json":
        schema = Formatter.create_schema()
    else:
        typer.echo(f"Don't know how to create {output.name}")
        raise typer.Exit(code=1)

    output.write_text(json.dumps(schema, indent=2) + "\n")


@app.command()
def format(
    output_path: Annotated[
        Path, typer.Argument(..., help="Directory to write output file(s) to")
    ],
    device_path: Annotated[
        Path, typer.Argument(..., help="Path to the .pvi.device.yaml file")
    ],
    formatter_path: Annotated[
        Path, typer.Argument(..., help="Path to the .pvi.formatter.yaml file")
    ],
    yaml_paths: Annotated[
        Optional[list[Path]],  # noqa
        typer.Option(
            ..., "--yaml-path", help="Paths to directories with .pvi.device.yaml files"
        ),
    ] = None,
):
    """Create screen product from device and formatter YAML"""
    yaml_paths = yaml_paths or []

    device = Device.deserialize(device_path)
    device.deserialize_parents(yaml_paths)

    formatter = Formatter.deserialize(formatter_path)
    formatter.format(device, output_path)


@app.command()
def generate_template(
    device_path: Annotated[
        Path, typer.Argument(..., help="Path to the .pvi.device.yaml file")
    ],
    pv_prefix: Annotated[str, typer.Argument(..., help="Prefix of PVI PV")],
    output_path: Annotated[Path, typer.Argument(..., help="Output file to generate")],
):
    """Create template with info tags for device signals"""
    device = Device.deserialize(device_path)
    format_template(device, pv_prefix, output_path)


@convert_app.command()
def device(
    output: Annotated[
        Path, typer.Argument(..., help="Directory to write output file to")
    ],
    header: Annotated[
        Optional[Path],  # noqa
        typer.Option(..., "--header", help="Path to the .h file to convert"),
    ] = None,
    templates: Annotated[
        Optional[list[Path]],  # noqa
        typer.Option(..., "--template", help="Paths to .template files to convert"),
    ] = None,
    name: Annotated[
        Optional[str],  # noqa
        typer.Option(
            ...,
            help=(
                "Name to use for Device. "
                "This is usually the same as the EPICS driver class."
            ),
        ),
    ] = None,
    label: Annotated[
        Optional[str],  # noqa
        typer.Option(
            ...,
            help=("Label for Device UI. Defaults to Device name."),
        ),
    ] = None,
    parent: Annotated[
        Optional[str],  # noqa
        typer.Option(..., help="Parent Device name."),
    ] = None,
):
    """Convert template to device YAML"""
    templates = templates or []

    device_name = None
    parent_name = None
    if header is not None:
        device_name, parent_name = extract_device_and_parent_class(header.read_text())
    if name is not None:
        device_name = name.replace(" ", "")  # No whitespace in file name
    if parent is not None:
        parent_name = parent
    if device_name is None:
        raise ValueError("Either Device name or header file must be provided.")

    component_group = TemplateConverter(templates).convert()
    device = Device(
        label=label or device_name, parent=parent_name, children=component_group
    )

    if not output.exists():
        os.mkdir(output)

    device.serialize(output / f"{device_name}.pvi.device.yaml")


@app.command()
def regroup(
    device_path: Annotated[
        Path, typer.Argument(..., help="Path to the device.yaml file to regroup")
    ],
    ui_paths: Annotated[
        list[Path],
        typer.Argument(..., help="Paths to the ui files to regroup the PVs by"),
    ],
):
    """Regroup a device.yaml file based on ui files that the PVs appear in"""
    device = Device.deserialize(device_path)
    device.children = group_by_ui(device, ui_paths)

    device.serialize(device_path)


@app.command()
def reconvert(
    device_path: Annotated[
        Path, typer.Argument(..., help="Path to the device.yaml file to add to")
    ],
    templates: Annotated[
        list[Path],
        typer.Option(..., "--template", help="Paths of templates to add PVs from"),
    ],
):
    """Add PVs to an existing device.yaml file based on extra / updated templates."""
    device = Device.deserialize(device_path)
    template_components = TemplateConverter(templates).convert()

    device.merge_components(template_components)

    device.serialize(device_path)


# test with: pipenv run python -m pvi
if __name__ == "__main__":
    app()



import re
from pathlib import Path

from pvi.device import (
    ComponentUnion,
    Device,
    Grid,
    Group,
    Tree,
    enforce_pascal_case,
    walk,
)


def find_pvs(pvs: list[str], file_path: Path) -> tuple[list[str], list[str]]:
    """Search for the PVs in the file and return lists of found and not found pvs"""

    with open(file_path) as f:
        file_content = f.read()

    pv_coordinates: dict[int, list[str]] = {}
    remaining_pvs = list(pvs)
    for pv in pvs:
        if pv not in file_content:
            continue

        # This regex searches for the first y coordinate before the given PV
        # https://regex101.com/r/41yX56/1
        widget_regex = re.compile(
            "".join(
                (
                    # The y-coordinate in a match group named `y`
                    r"y=(?P<y>\d+)",
                    # Negative lookahead asserting no further y-coordinate before PV
                    r"(?!.*y=.*" + pv + r")",
                    # Any characters and then PV
                    r".*" + pv,
                )
            ),
            re.MULTILINE | re.DOTALL,
        )
        match = re.search(widget_regex, file_content)

        assert match, f"{pv} found in {file_path.name} but did not match a y coordinate"

        y = int(match["y"])
        if y in pv_coordinates:
            pv_coordinates[y].append(pv)
        else:
            pv_coordinates[y] = [pv]

        remaining_pvs.remove(pv)

    grouped_pvs: list[str] = []
    for coord in sorted(pv_coordinates.keys()):
        grouped_pvs.extend(pv_coordinates[coord])

    return grouped_pvs, remaining_pvs


def group_by_ui(device: Device, ui_paths: list[Path]) -> Tree:
    signals: list[ComponentUnion] = list(walk(device.children))

    # PVs without macros to search for in UI
    pv_names = [s.name for s in signals]

    group_pv_map: dict[str, list[str]] = {}
    for ui in ui_paths:
        ui_pvs, pv_names = find_pvs(pv_names, ui)
        if ui_pvs:
            group_pv_map[ui.stem] = ui_pvs

    if pv_names:
        print(f"Did not find group for {' | '.join(pv_names)}")

    # Create groups for parameters we found in the files
    ui_groups: list[Group] = [
        Group(
            name=enforce_pascal_case(group_name),
            layout=Grid(labelled=True),
            children=[  # Note: Need to preserve order in group_pvs here
                signal
                for pv_name in group_pvs
                for signal in signals
                if signal.name == pv_name
            ],
            label=group_name if enforce_pascal_case(group_name) != group_name else None,
        )
        for group_name, group_pvs in group_pv_map.items()
    ]

    # Separate any parameters we failed to find a group for
    grouped_pvs = [pv_name for group in ui_groups for pv_name in group.children]
    ungrouped_pvs = [pv_name for pv_name in signals if pv_name not in grouped_pvs]

    # Add any ungrouped parameters on the end
    if ungrouped_pvs:
        ui_groups.append(
            Group(
                name=enforce_pascal_case(device.label) + "Misc",
                layout=Grid(labelled=True),
                children=ungrouped_pvs,
                label=device.label + " Ungrouped",
            )
        )

    return ui_groups



from __future__ import annotations

import re
from typing import Any

from pvi._format.utils import Bounds, split_with_sep
from pvi._format.widget import UITemplate, WidgetFormatter
from pvi.device import TextFormat, TextRead, TextWrite, WidgetUnion

EDL_TEXT_FORMATS = {
    TextFormat.decimal: "decimal",
    TextFormat.hexadecimal: "hex",
    TextFormat.engineer: "engineer",
    TextFormat.exponential: "exp",
    TextFormat.string: "default",
}


class EdlTemplate(UITemplate[str]):
    def __init__(self, text: str):
        assert "endGroup" not in text, "Can't do groups"
        self.screen, text = split_with_sep(text, "\nendScreenProperties\n", 1)
        self.widgets = split_with_sep(text, "\nendObjectProperties\n")

    def set(
        self,
        template: str,
        bounds: Bounds | None = None,
        widget: WidgetUnion | None = None,
        properties: dict[str, Any] | None = None,
    ) -> str:
        properties = properties or {}

        if bounds:
            for k in "xywh":
                properties[k] = getattr(bounds, k)
        for item, value in properties.items():
            if item == "displayFileName":
                value = f"0 {value}"  # These are items in an array but we only use one

            multiline = re.compile(rf"^{item} {{[^}}]*}}$", re.MULTILINE | re.DOTALL)
            if multiline.search(template):
                pattern = multiline
                lines = str(value).splitlines()
                value = "\n".join(["{"] + [f'  "{x}"' for x in lines] + ["}"])
            else:
                # Single line
                pattern = re.compile(rf"^{item} .*$", re.MULTILINE)
                if isinstance(value, str):
                    value = f'"{value}"'

            template, n = pattern.subn(f"{item} {value}", template)
            assert n == 1, f"No replacements made for {item}"

        # Add additional properties from widget
        match widget:
            case TextWrite(format=format) | TextRead(format=format) if (
                is_text_widget(template) and format is not None
            ):
                template = add_property(
                    template, "displayMode", EDL_TEXT_FORMATS[TextFormat(format)]
                )
            case _:
                pass

        return template

    def search(self, search: str) -> str:
        matches = [t for t in self.widgets if re.search(search, t)]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"
        return matches[0]

    def create_group(
        self,
        group_object: list[str],
        children: list[WidgetFormatter[str]],
        padding: Bounds | None = None,
    ) -> list[str]:
        padding = padding or Bounds()

        texts: list[str] = []

        for c in children:
            c.bounds.x += padding.x
            c.bounds.y += padding.y
            texts += c.format()

        return group_object + texts


def is_text_widget(text: str):
    return text.startswith("\n# (Text")


def add_property(text: str, property: str, value: str) -> str:
    end = "endObjectProperties\n"
    return text.replace(end, f'{property} "{value}"\n{end}')



import os
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

from pvi import __version__
from pvi.__main__ import app
from pvi._format import Formatter
from pvi.device import Device, Grid, Group

HERE = Path(__file__).parent


def test_cli_version():
    cmd = [sys.executable, "-m", "pvi", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


@pytest.mark.parametrize(
    "filename",
    [
        "pvi.device.schema.json",
        "pvi.formatter.schema.json",
    ],
)
def test_schemas(tmp_path, helper, filename):
    expected_path = HERE.parent / "schemas" / filename
    helper.assert_cli_output_matches(app, expected_path, "schema", tmp_path / filename)


@pytest.fixture(scope="module")
def formats_to_skip():
    return {
        ".adl": ["BitField", "ArrayTrace", "ImageRead", "TableRead", "TableWrite"],
        ".edl": ["ArrayTrace", "ImageRead", "TableRead", "TableWrite"],
        ".bob": [],
    }


@pytest.mark.parametrize(
    "widget_name",
    [
        "ArrayTrace",
        "BitField",
        "ButtonPanel",
        "CheckBox",
        "ComboBox",
        "ImageRead",
        "LED",
        "ProgressBar",
        "TableRead",
        "TableWrite",
        "TextRead",
        "TextWrite",
        "ToggleButton",
    ],
)
@pytest.mark.parametrize(
    "formatter,format",
    [
        ("aps.adl.pvi.formatter.yaml", ".adl"),
        ("dls.edl.pvi.formatter.yaml", ".edl"),
        ("dls.bob.pvi.formatter.yaml", ".bob"),
    ],
)
def test_format(tmp_path, helper, formatter, format, widget_name, formats_to_skip):
    if widget_name in formats_to_skip[format]:
        pytest.skip(f"{widget_name} not supported in {format} format")
    filename = widget_name + format
    expected_path = HERE / "format" / "output" / "all_widgets" / filename
    input_path = HERE / "format" / "input"
    formatter_path = input_path / formatter
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / filename,
        input_path / "all_widgets" / (widget_name + ".pvi.device.yaml"),
        formatter_path,
    )


def test_format_parent_child(tmp_path, helper):
    expected_path = HERE / "format" / "output" / "parent_child.bob"
    input_path = HERE / "format" / "input"
    formatter_path = input_path / "dls.bob.pvi.formatter.yaml"
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / "parent_child.bob",
        HERE / "format" / "input" / "child.pvi.device.yaml",
        formatter_path,
    )


def test_signal_default_widgets(tmp_path, helper):
    expected_path = HERE / "format" / "output" / "signal_default_widgets.bob"
    input_path = HERE / "format" / "input"
    formatter_path = input_path / "dls.bob.pvi.formatter.yaml"
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / "signal_default_widgets.bob",
        HERE / "format" / "input" / "signal_default_widgets.pvi.device.yaml",
        formatter_path,
    )


def test_convert_header(tmp_path, helper):
    expected_path = HERE / "convert" / "output"
    input_path = HERE / "convert" / "input"
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "convert device",
        tmp_path,
        "--header",
        input_path / "simDetector.h",
        "--template",
        input_path / "simDetector.template",
    )


def test_convert_device_name(tmp_path, helper):
    expected_path = HERE / "convert" / "output"
    input_path = HERE / "convert" / "input"
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "convert device",
        tmp_path,
        "--name",
        "Mako125B",
        "--label",
        "GenICam Mako125B",
        "--parent",
        "GenICamDriver",
        "--template",
        input_path / "Mako125B.template",
    )


def test_reconvert(tmp_path, helper):
    expected_path = HERE / "reconvert" / "output"
    input_path = HERE / "reconvert" / "input"
    # Make a copy to modify in place
    shutil.copy(
        input_path / "simDetector.pvi.device.yaml",
        tmp_path / "simDetector.pvi.device.yaml",
    )
    helper.assert_cli_output_matches(
        app,
        expected_path / "simDetector.pvi.device.yaml",
        "reconvert",
        tmp_path / "simDetector.pvi.device.yaml",
        "--template",
        input_path / "simDetectorExtra.template",
        "--template",
        input_path / "simDetector.template",
    )


@pytest.mark.parametrize(
    "input_yaml,formatter,output",
    [
        (
            "static_table.pvi.device.yaml",
            "dls.edl.pvi.formatter.yaml",
            "static_table.edl",
        ),
        (
            "static_table.pvi.device.yaml",
            "dls.bob.pvi.formatter.yaml",
            "static_table.bob",
        ),
        (
            "static_table.pvi.device.yaml",
            "aps.pvi.formatter.yaml",
            "static_table.adl",
        ),
    ],
)
def test_static_table(tmp_path, helper, input_yaml, formatter, output):
    expected_path = HERE / "format" / "output" / output
    input_path = HERE / "format" / "input"
    formatter_path = HERE / "../formatters/" / formatter
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / output,
        input_path / input_yaml,
        formatter_path,
    )


@pytest.mark.parametrize(
    "formatter,format,skip",
    [
        (
            "aps.adl.pvi.formatter.yaml",
            ".adl",
            ["ArrayTrace", "ImageRead", "TableRead", "TableWrite", "BitField"],
        ),
        (
            "dls.edl.pvi.formatter.yaml",
            ".edl",
            ["ArrayTrace", "ImageRead", "TableRead", "TableWrite"],
        ),
        ("dls.bob.pvi.formatter.yaml", ".bob", []),
    ],
)
def test_combine_widgets(tmp_path, helper, formatter, format, skip):
    children = []
    for path in sorted((HERE / "format" / "input" / "all_widgets").iterdir()):
        device = Device.deserialize(path)
        children.append(
            Group(
                name=path.name.split(".")[0],
                label=device.label,
                children=[
                    child
                    for child in device.children
                    if not any(s in child.name for s in skip)
                ],
                layout=Grid(),
            )
        )
    all_device = Device(label="All Widgets", children=children)

    screen_filename = f"all_widgets{format}"
    output_path = tmp_path / screen_filename
    expected_path = HERE / "format" / "output" / "all_widgets" / screen_filename
    my_formatter = Formatter.deserialize(HERE / "format" / "input" / formatter)
    my_formatter.format(all_device, output_path)

    if os.environ.get("PVI_REGENERATE_OUTPUT", None):
        shutil.copy(output_path, expected_path)

    helper.assert_output_matches(expected_path, output_path)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Any, Generic, Self, TypeVar

from pvi._format.utils import Bounds
from pvi.device import (
    LED,
    ArrayTrace,
    BitField,
    ButtonPanel,
    CheckBox,
    ComboBox,
    Group,
    ImageRead,
    ProgressBar,
    TableRead,
    TableWrite,
    TextRead,
    TextWrite,
    ToggleButton,
    WidgetUnion,
)

T = TypeVar("T")


class UITemplate(Generic[T]):
    screen: T

    def search(self, search: str) -> T:
        """Extract a snippet from the template

        Args:
            search: The search expression
                This must be unique in the template so that there is only one match

        Returns:
            The snippet matching the search

        """
        raise NotImplementedError(self)

    def set(
        self,
        template: T,
        bounds: Bounds | None = None,
        widget: WidgetUnion | None = None,
        properties: dict[str, Any] | None = None,
    ) -> T:
        """Modify template elements with component data

        Args:
            template: A template element
            bounds: The size and position of the widget
            widget: A Widget to define some additional properties for the element
            properties: The element properties in the form: {placeholder: value}
                placeholder is value to find in the template and value is the component
                value to replace it with

        Returns:
            The modified element

        """
        raise NotImplementedError(self)

    def create_group(
        self,
        group_object: list[T],
        children: list[WidgetFormatter[T]],
        padding: Bounds,
    ) -> list[T]:
        """Create a group widget with its children embedded and appropriately padded"""
        raise NotImplementedError(self)


@dataclass
class WidgetFormatter(Generic[T]):
    bounds: Bounds

    def format(self) -> list[T]:
        """Instances should be created using `from_template`, which defines `format`"""
        raise NotImplementedError(self)

    @classmethod
    def from_template(
        cls: type[Self],
        template: UITemplate[T],
        search: str,
        sized: Callable[[Bounds], Bounds] = Bounds.clone,
        widget_formatter_hook: Callable[[Bounds, str], list[WidgetFormatter[T]]]
        | None = None,
        property_map: dict[str, str] | None = None,
    ) -> type[Self]:
        """Create a WidgetFormatter class from the given template

        Create a `format` method that searches the template for the `search` section and
        extract it as a formattable template for a single widget. Create a child class
        of `parent_cls` implementing the method.

        Args:
            parent_cls: The class of formatter to create
            template: UI template to search for formattable widget sections
            search: Search expression for the widget section to find
            sized: Function to pad or restrict the `Bounds` of the formatted widget
            widget_formatter_hook: Hook to add additional formatters.
                Currently this is unused and available for child classes to use in
                overrided `from_template` calls.
            property_map: Map of template macro string to `WidgetFormatter` property.
                Instances of the macro will be replaced with the value of the property.
        """

        def format(self: WidgetFormatter[T]) -> list[T]:
            properties: dict[str, str] = {}
            if property_map is not None:
                for placeholder, widget_property in property_map.items():
                    assert hasattr(self, widget_property), (
                        f"{self} has no property {widget_property}"
                    )
                    properties[placeholder] = getattr(self, widget_property)

            return [
                template.set(
                    template.search(search),
                    sized(self.bounds),
                    self.widget if isinstance(self, PVWidgetFormatter) else None,
                    properties,
                )
            ]

        return type(  # type: ignore
            f"""{cls.__name__}<{search.strip('"')}>""",
            (cls,),
            {"format": format},
        )


@dataclass
class LabelWidgetFormatter(WidgetFormatter[T]):
    text: str
    tooltip: str = ""


@dataclass
class PVWidgetFormatter(WidgetFormatter[T]):
    pv: str
    widget: WidgetUnion


@dataclass
class ActionWidgetFormatter(WidgetFormatter[T]):
    label: str
    pv: str
    value: str

    @property
    def tooltip(self) -> str:
        return f"{self.pv} = {self.value}"


@dataclass
class SubScreenWidgetFormatter(WidgetFormatter[T]):
    label: str
    file_name: str
    components: Group | None = None
    macros: dict[str, str] = field(default_factory=dict[str, str])


class GroupType(StrEnum):
    GROUP = "GROUP"
    SCREEN = "SCREEN"


@dataclass
class GroupFormatter(WidgetFormatter[T]):
    bounds: Bounds
    title: str
    children: list[WidgetFormatter[T]]

    def format(self) -> list[T]:
        """Instances should be created using `from_template`, which defines `format`"""
        raise NotImplementedError(self)

    def __post_init__(self) -> None:
        self.resize()

    def resize(self):
        pass

    @classmethod
    def from_template(
        cls: type[Self],
        template: UITemplate[T],
        search: str,
        sized: Callable[[Bounds], Bounds] = Bounds.clone,
        widget_formatter_hook: Callable[[Bounds, str], list[WidgetFormatter[T]]]
        | None = None,
        property_map: dict[str, str] | None = None,
    ) -> type[Self]:
        """Create a WidgetFormatter class from the given template

        Create a `format` method that searches the template for the `search` section and
        extract it as a formattable template for a single widget. Create a child class
        of `parent_cls` implementing the method.

        Args:
            parent_cls: The class of formatter to create
            template: UI template to search for formattable widget sections
            search: Search expression for the widget section to find
            sized: Function to pad or restrict the `Bounds` of the formatted widget
            widget_formatter_hook: Callable to format widgets for the group itself, such
                as a label or a box around the group components.
            property_map: Map of template macro string to `WidgetFormatter` property.
                Instances of the macro will be replaced with the value of the property.
        """

        def format(self: GroupFormatter[T]) -> list[T]:
            padding = sized(self.bounds)
            texts: list[T] = []
            made_widgets: list[T] = []

            if search == GroupType.SCREEN:
                properties: dict[str, str] = {}
                if property_map is not None:
                    for placeholder, widget_property in property_map.items():
                        assert hasattr(self, widget_property), (
                            f"{self} has no property {widget_property}"
                        )
                        properties[placeholder] = getattr(self, widget_property)

                texts.append(
                    template.set(template.screen, self.bounds, properties=properties)
                )
                # Make screen title
                if widget_formatter_hook:
                    for widget in widget_formatter_hook(self.bounds, self.title):
                        texts += widget.format()
                for c in self.children:
                    c.bounds.x += padding.x
                    c.bounds.y += padding.y
                    texts += c.format()

            if search == GroupType.GROUP:
                # Make group object
                if widget_formatter_hook:
                    for widget in widget_formatter_hook(self.bounds, self.title):
                        made_widgets += widget.format()
                texts += template.create_group(made_widgets, self.children, padding)
            return texts

        def resize(self: GroupFormatter[T]):
            """Resize based on widget template.

            Called in __init__ by parent class.

            """
            padding = sized(self.bounds)
            self.bounds = Bounds(
                x=self.bounds.x, y=self.bounds.y, w=padding.w, h=padding.h
            )
            pass

        return type(  # type: ignore
            f"{cls.__name__}<{search}>",
            (cls,),
            {"format": format, "resize": resize},
        )


@dataclass
class WidgetFormatterFactory(Generic[T]):
    header_formatter_cls: type[LabelWidgetFormatter[T]]
    label_formatter_cls: type[LabelWidgetFormatter[T]]
    action_formatter_cls: type[ActionWidgetFormatter[T]]
    sub_screen_formatter_cls: type[SubScreenWidgetFormatter[T]]
    led_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    progress_bar_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    text_read_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    check_box_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    toggle_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    combo_box_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    text_write_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    table_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    bitfield_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    array_trace_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    button_panel_formatter_cls: type[PVWidgetFormatter[T]] | None = None
    image_read_formatter_cls: type[PVWidgetFormatter[T]] | None = None

    def pv_widget_formatter(
        self,
        widget: WidgetUnion,
        bounds: Bounds,
        pv: str,
    ) -> PVWidgetFormatter[T]:
        """Convert a component into its WidgetFormatter equivalent

        Args:
            widget: The read/write widget property of a component
            bounds: The size and position of the widget (x,y,w,h).
            pv: The process variable assigned to a component

        Returns:
            A WidgetFormatter representing the component
        """

        widget_formatter_classes: dict[type, type[PVWidgetFormatter[T]] | None] = {
            # Currently supported formatters of ReadWidget/WriteWidget Components
            LED: self.led_formatter_cls,
            ProgressBar: self.progress_bar_formatter_cls,
            TextRead: self.text_read_formatter_cls,
            TableRead: self.table_formatter_cls,
            CheckBox: self.check_box_formatter_cls,
            ToggleButton: self.toggle_formatter_cls,
            ComboBox: self.combo_box_formatter_cls,
            TextWrite: self.text_write_formatter_cls,
            TableWrite: self.table_formatter_cls,
            BitField: self.bitfield_formatter_cls,
            ArrayTrace: self.array_trace_formatter_cls,
            ButtonPanel: self.button_panel_formatter_cls,
            ImageRead: self.image_read_formatter_cls,
        }
        if isinstance(widget, TextRead | TextWrite):
            bounds.h *= widget.get_lines()

        widget_formatter_cls = widget_formatter_classes[type(widget)]
        if widget_formatter_cls is None:
            raise NotImplementedError(
                f"No formatter provided for widget {type(widget).__name__}"
            )
        return widget_formatter_cls(bounds=bounds, pv=pv, widget=widget)


def max_x(widgets: list[WidgetFormatter[T]]) -> int:
    """Given multiple widgets, calulate the maximum x position that they occupy"""
    if widgets:
        return max(w.bounds.x + w.bounds.w for w in widgets)
    else:
        return 0


def max_y(widgets: list[WidgetFormatter[T]]) -> int:
    """Given multiple widgets, calulate the maximum y position that they occupy"""
    if widgets:
        return max(w.bounds.y + w.bounds.h for w in widgets)
    else:
        return 0


def next_x(widgets: list[WidgetFormatter[T]], spacing: int = 0) -> int:
    """Given multiple widgets, calulate the next feasible location for an
    additional widget in the x axis"""
    if widgets:
        return max_x(widgets) + spacing
    else:
        return 0


def next_y(widgets: list[WidgetFormatter[T]], spacing: int = 0) -> int:
    """Given multiple widgets, calulate the next feasible location for an
    additional widget in the y axis"""
    if widgets:
        return max_y(widgets) + spacing
    else:
        return 0



from pathlib import Path
from typing import Annotated

from lxml.etree import (
    _Element,  # pyright: ignore [reportPrivateUsage]
    fromstring,
    tostring,
)
from pydantic import Field

from pvi._format.bob import BobTemplate, find_element
from pvi._format.edl import EdlTemplate
from pvi._format.screen import (
    ScreenFormatterFactory,
    ScreenLayout,
    WidgetFormatterFactory,
)
from pvi._format.widget import (
    ActionWidgetFormatter,
    GroupFormatter,
    GroupType,
    LabelWidgetFormatter,
    PVWidgetFormatter,
    SubScreenWidgetFormatter,
    WidgetFormatter,
)
from pvi.device import Device

from .base import Formatter
from .utils import Bounds, with_title


class DLSFormatter(Formatter):
    spacing: Annotated[int, Field(description="Spacing between widgets")] = 5
    title_height: Annotated[int, Field(description="Height of screen title bar")] = 25
    max_height: Annotated[int, Field(description="Max height of the screen")] = 900
    label_width: Annotated[
        int, Field(description="Width of the widget description labels")
    ] = 150
    widget_width: Annotated[int, Field(description="Width of the widgets")] = 200
    widget_height: Annotated[int, Field(description="Height of the widgets")] = 20

    def format(self, device: Device, path: Path) -> None:
        if path.suffix == ".edl":
            f = self.format_edl
        elif path.suffix == ".bob":
            f = self.format_bob
        else:
            raise ValueError("Can only write .edl or .bob files")
        f(device, path)

    def format_edl(self, device: Device, path: Path):
        template = EdlTemplate((Path(__file__).parent / "dls.edl").read_text())
        screen_layout = ScreenLayout(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=10,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=5,
            group_width_offset=0,
        )
        widget_formatter_factory = WidgetFormatterFactory[str](
            header_formatter_cls=LabelWidgetFormatter[str].from_template(
                template,
                search='"Heading"',
                property_map={"value": "text"},
            ),
            label_formatter_cls=LabelWidgetFormatter[str].from_template(
                template,
                search='"Label"',
                property_map={"value": "text"},
            ),
            led_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"LED"',
                sized=Bounds.square,
                property_map={"controlPv": "pv"},
            ),
            progress_bar_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ProgressBar"',
                property_map={"indicatorPv": "pv"},
            ),
            text_read_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"TextRead"',
                property_map={"controlPv": "pv"},
            ),
            check_box_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ComboBox"',
                property_map={"controlPv": "pv"},
            ),
            toggle_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ToggleButton"',
                property_map={"controlPv": "pv"},
            ),
            combo_box_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"ComboBox"',
                property_map={"controlPv": "pv"},
            ),
            text_write_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"TextWrite"',
                property_map={"controlPv": "pv"},
            ),
            action_formatter_cls=ActionWidgetFormatter[str].from_template(
                template,
                search='"SignalX"',
                property_map={
                    "onLabel": "label",
                    "offLabel": "label",
                    "controlPv": "pv",
                },
            ),
            sub_screen_formatter_cls=SubScreenWidgetFormatter[str].from_template(
                template,
                search='"SubScreenFile"',
                property_map={"displayFileName": "file_name"},
            ),
            bitfield_formatter_cls=PVWidgetFormatter[str].from_template(
                template,
                search='"LED"',
                property_map={"controlPv": "pv"},
            ),
        )
        screen_title_cls = LabelWidgetFormatter[str].from_template(
            template,
            search='"Title"',
            property_map={"value": "text"},
        )
        group_title_cls = LabelWidgetFormatter[str].from_template(
            template,
            search='"  Group  "',
            property_map={"value": "text"},
        )
        group_box_cls = WidgetFormatter[str].from_template(
            template, search="fillColor index 5"
        )

        def create_group_box_formatter(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[str]]:
            x, y, w, h = bounds.x, bounds.y, bounds.w, bounds.h
            return [
                group_box_cls(
                    bounds=Bounds(
                        x=x,
                        y=y + screen_layout.spacing,
                        w=w,
                        h=h - screen_layout.spacing,
                    )
                ),
                group_title_cls(
                    bounds=Bounds(x=x, y=y, w=w, h=screen_layout.group_label_height),
                    text=f"  {title}  ",
                ),
            ]

        def create_screen_title_formatter(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[str]]:
            return [
                screen_title_cls(
                    bounds=Bounds(x=0, y=0, w=bounds.w, h=screen_layout.title_height),
                    text=title,
                )
            ]

        formatter_factory: ScreenFormatterFactory[str] = ScreenFormatterFactory(
            screen_formatter_cls=GroupFormatter[str].from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(screen_layout.spacing, screen_layout.title_height),
                widget_formatter_hook=create_screen_title_formatter,
            ),
            group_formatter_cls=GroupFormatter[str].from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    screen_layout.spacing, screen_layout.group_label_height
                ),
                widget_formatter_hook=create_group_box_formatter,
            ),
            widget_formatter_factory=widget_formatter_factory,
            layout=screen_layout,
            base_file_name=path.stem,
        )
        title = f"{device.label}"

        screen_formatter, sub_screens = formatter_factory.create_screen_formatter(
            device.children, title
        )

        path.write_text("".join(screen_formatter.format()))
        for sub_screen_name, sub_screen_formatter in sub_screens:
            sub_screen_path = Path(path.parent / f"{sub_screen_name}{path.suffix}")
            sub_screen_path.write_text("".join(sub_screen_formatter.format()))

    def format_bob(self, device: Device, path: Path):
        template = BobTemplate(str(Path(__file__).parent / "dls.bob"))
        # LP DOCS REF: Define the layout properties
        screen_layout = ScreenLayout(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=26,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=18,
            group_width_offset=26,
        )
        # SW DOCS REF: Extract widget types from template file
        widget_formatter_factory: WidgetFormatterFactory[_Element] = (
            WidgetFormatterFactory(
                header_formatter_cls=LabelWidgetFormatter[_Element].from_template(
                    template,
                    search="Heading",
                    property_map={"text": "text"},
                ),
                label_formatter_cls=LabelWidgetFormatter[_Element].from_template(
                    template,
                    search="Label",
                    property_map={"text": "text", "tooltip": "tooltip"},
                ),
                led_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="LED",
                    sized=Bounds.square,
                    property_map={"pv_name": "pv"},
                ),
                progress_bar_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ProgressBar",
                    property_map={"pv_name": "pv"},
                ),
                bitfield_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="BitField",
                    property_map={"pv_name": "pv"},
                ),
                button_panel_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ButtonPanel",
                    property_map={"pv_name": "pv"},
                ),
                array_trace_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ArrayTrace",
                    property_map={"y_pv": "pv"},
                ),
                image_read_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ImageRead",
                    property_map={"pv_name": "pv"},
                ),
                text_read_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="TextUpdate",
                    property_map={"pv_name": "pv"},
                ),
                check_box_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="CheckBox",
                    property_map={"pv_name": "pv"},
                ),
                toggle_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ToggleButton",
                    sized=Bounds.square,
                    property_map={"pv_name": "pv"},
                ),
                combo_box_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="ComboBox",
                    property_map={"pv_name": "pv"},
                ),
                text_write_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="TextEntry",
                    property_map={"pv_name": "pv"},
                ),
                table_formatter_cls=PVWidgetFormatter[_Element].from_template(
                    template,
                    search="Table",
                    property_map={"pv_name": "pv"},
                ),
                action_formatter_cls=ActionWidgetFormatter[_Element].from_template(
                    template,
                    search="WritePV",
                    property_map={
                        "text": "label",
                        "pv_name": "pv",
                        "value": "value",
                        "tooltip": "tooltip",
                    },
                ),
                sub_screen_formatter_cls=SubScreenWidgetFormatter[
                    _Element
                ].from_template(
                    template,
                    search="OpenDisplay",
                    property_map={
                        "file": "file_name",
                        "text": "label",
                        "macros": "macros",
                    },
                ),
            )
        )
        # MAKE_WIDGETS DOCS REF: Define screen and group widgets
        screen_title_cls = LabelWidgetFormatter[_Element].from_template(
            template,
            search="Title",
            property_map={"text": "text"},
        )
        group_title_cls = LabelWidgetFormatter[_Element].from_template(
            template,
            search="Group",
            property_map={"name": "text"},
        )

        def create_group_object_formatter(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[_Element]]:
            return [
                group_title_cls(
                    bounds=Bounds(x=bounds.x, y=bounds.y, w=bounds.w, h=bounds.h),
                    text=f"{title}",
                )
            ]

        def create_screen_title_formatter(
            bounds: Bounds, title: str
        ) -> list[WidgetFormatter[_Element]]:
            return [
                screen_title_cls(
                    bounds=Bounds(x=0, y=0, w=bounds.w, h=screen_layout.title_height),
                    text=title,
                )
            ]

        # SCREEN_INI DOCS REF: Construct a screen object
        formatter_factory: ScreenFormatterFactory[_Element] = ScreenFormatterFactory(
            screen_formatter_cls=GroupFormatter[_Element].from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(screen_layout.spacing, screen_layout.title_height),
                widget_formatter_hook=create_screen_title_formatter,
            ),
            group_formatter_cls=GroupFormatter[_Element].from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    screen_layout.spacing, screen_layout.group_label_height
                ),
                widget_formatter_hook=create_group_object_formatter,
            ),
            widget_formatter_factory=widget_formatter_factory,
            layout=screen_layout,
            base_file_name=path.stem,
        )
        # SCREEN_FORMAT DOCS REF: Format the screen
        title = f"{device.label}"

        screen_formatter, sub_screens = formatter_factory.create_screen_formatter(
            device.children, title
        )

        write_bob(screen_formatter, path)
        for sub_screen_name, sub_screen_formatter in sub_screens:
            sub_screen_path = Path(path.parent / f"{sub_screen_name}{path.suffix}")
            write_bob(sub_screen_formatter, sub_screen_path)


# SCREEN_WRITE DOCS REF: Generate the screen file


def write_bob(screen_formatter: GroupFormatter[_Element], path: Path):
    # The root:'Display' is always the first element in texts
    texts = screen_formatter.format()
    element_tree = fromstring(tostring(texts[0]), None)
    for element in texts[:0:-1]:
        grid_step_y = element_tree.find("grid_step_y")
        if grid_step_y is None:
            raise ValueError(f"Could not find grid_step_y in element {element}")

        element_tree.insert(element_tree.index(grid_step_y) + 1, element)

    element_tree = element_tree.getroottree()
    find_element(element_tree, "name").text = screen_formatter.title  # type: ignore
    element_tree.write(str(path), pretty_print=True)



import re


def extract_device_and_parent_class(header_text: str) -> tuple[str, str]:
    # e.g. extract 'NDPluginDriver' and 'asynNDArrayDriver' from
    # class epicsShareClass NDPluginDriver : public asynNDArrayDriver, public epicsThreadRunable {  # noqa
    class_extractor = re.compile(r"class.*\s+(\w+)\s+:\s+\w+\s+(\w+).*")
    match = re.search(class_extractor, header_text)
    assert match, "Can't find device class and parent class in header file"
    classname, parent = match.groups()
    return classname, parent


def extract_define_strs(header_text: str, info_strings: list[str]) -> list[str]:
    # e.g. extract: #define SimGainXString                "SIM_GAIN_X";
    define_extractor = re.compile(r'\#define[_A-Za-z0-9 ]*"[^"]*".*')
    definitions = re.findall(define_extractor, header_text)
    # We only want to extract the defines for the given parameter infos
    definitions = filter_strings(definitions, info_strings)
    return definitions


def extract_create_param_strs(source_text: str, param_strings: list[str]) -> list[str]:
    # e.g. extract: createParam(SimGainXString, asynParamFloat64, &SimGainX);
    create_param_extractor = re.compile(r"((?:this->)?createParam\([^\)]*\);.*)")
    create_param_strs = re.findall(create_param_extractor, source_text)
    # We only want to extract the createParam calls for the given parameter strings
    create_param_strs = filter_strings(create_param_strs, param_strings)
    return create_param_strs


def extract_index_declarations(header_text: str, index_names: list[str]) -> list[str]:
    # e.g. extract:     int SimGainX;
    declaration_extractor = re.compile(r"\s*int [^;]*;")
    declarations = re.findall(declaration_extractor, header_text)
    # We only want to extract the declarations for the given index names - this also
    # filters any generic int parameter definitions in the class and some comments
    declarations = filter_strings(declarations, index_names)
    return declarations


def parse_definition_str(definition_str: str) -> tuple[str, str]:
    # e.g. from: #define SimGainXString                "SIM_GAIN_X";
    # extract:
    # Group1: SimGainXString
    # Group2: SIM_GAIN_X
    define_extractor = re.compile(r'(?:\#define) (\w+) *"([^"]*)')
    string_info_pair = re.findall(define_extractor, definition_str)[0]
    return string_info_pair


def parse_create_param_str(create_param_str: str) -> tuple[str, str]:
    # e.g. from: createParam(SimGainXString, asynParamFloat64, &SimGainX);
    # extract: SimGainXString,               asynParamFloat64, &SimGainX
    create_param_extractor = re.compile(r"(?:createParam\()([^\)]*)(?:\))")
    # There are two signatures for createParam. 3 argument and 4 argument
    # I think we are only ever interested in the final three arguments
    create_param_args = (
        re.findall(create_param_extractor, create_param_str)[0]
        .replace(" ", "")
        .split(",")[-3:]
    )
    string_index_pair = (create_param_args[0], create_param_args[2].strip("&"))
    return string_index_pair


def insert_param_set_accessors(source_text: str, parameters: list[str]) -> str:
    for parameter in parameters:
        # Only match parameter name exactly, not others with same prefix
        source_text = re.sub(
            r"(\W)" + parameter + r"(\W)",
            r"\1paramSet->" + parameter + r"\2",
            source_text,
        )
    return source_text


def filter_strings(strings: list[str], filters: list[str]) -> list[str]:
    return [
        string for string in strings if any(filter_ in string for filter_ in filters)
    ]


def get_param_set(driver: str):
    return "asynParamSet" if driver == "asynPortDriver" else driver + "ParamSet"



from __future__ import annotations

from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
    Union,
    get_args,
)

from pydantic import BaseModel, ConfigDict, Discriminator, Field, Tag, computed_field
from pydantic.fields import FieldInfo
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
)


class TypedModel(BaseModel):
    """A Base class for members of tagged unions discriminated by the class name.

    This class defines some hooks called by pydantic during validation and schema
    generation.

    Child classes will automatically have a type field defined, which can be used as a
    discriminator when constructing tagged unions using pydantic `Discriminator` and
    `Tag`.

    """

    # Do not allow extra fields during validation
    model_config = ConfigDict(extra="forbid")

    # Whether child models have been rebuilt with type field inserted
    models_typed: ClassVar[bool] = False

    @computed_field  # type: ignore
    @property
    def type(self) -> str:
        """Property to create type field from class name when serializing."""
        return self.__class__.__name__

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        """Add computed type field to model_fields to include it in the schema.

        The `model_rebuild` method must be called to propagate this to the core schema
        of the models, once all `TypedModel` child classes have been defined. The
        `model_json_schema` method is overridden in this class to make sure this is
        applied automatically before generating schema.

        """
        super().__pydantic_init_subclass__(**kwargs)

        cls.model_fields["type"] = FieldInfo(
            annotation=Literal[cls.__name__],  # type: ignore
            default=cls.__name__,
        )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchema,
        mode: JsonSchemaMode = "validation",
    ):
        """Ensure all child models have type field added before generating schema."""
        if not cls.models_typed:
            TypedModel.rebuild_child_models()

        return super().model_json_schema(by_alias, ref_template, schema_generator, mode)

    @classmethod
    def rebuild_child_models(cls):
        """Recursively rebuild all subclass models to add type into core schema."""
        for subclass in cls.__subclasses__():
            subclass.model_rebuild(force=True)
            subclass.rebuild_child_models()

    @classmethod
    def _tag(cls):
        """Create a pydantic `Tag` for this class to include it in tagged unions."""
        return Annotated[cls, Tag(cls.__name__)]

    @staticmethod
    def discriminator():
        """Create a pydantic `Discriminator` for a tagged union of `TypedModel`."""
        return Field(discriminator=Discriminator(TypedModel._get_type_name))

    @staticmethod
    def _get_type_name(x: TypedModel | dict[str, Any]) -> str | None:
        """Get the type name from an instance or serialized `dict` of `TypeModel`.

        This is a callable for pydantic Discriminator to discriminate between types in a
        tagged union of `TypedModel` child classes.

        If given an instance of `TypedModel` then this method is being called to
        serialize an instance. The type field of the entry for this instance should be
        its class name.

        If given a dictionary, then an instance is being deserialized. The name of the
        class to be instantiated is given by the type field, and the remaining fields
        should be passed as fields to the class.

        In any other case, return `None` to cause a pydantic validation error.

        Args:
            x: `TypedModel` instance or serialized `dict` of a `TypedModel`

        """
        match x:
            case TypedModel():
                return x.__class__.__name__
            case dict() as serialized:
                return serialized.pop("type", None)


def as_tagged_union(union: type[TypedModel]):
    """Create a tagged union from a `Union` of `TypedModel`.

    Members will be tagged with their class name to be discriminated by pydantic.

    Args:
        union: `Union` of `TypedModel` to convert to a tagged union

    """
    union_members = get_args(union)

    return Annotated[
        Union[tuple(cls._tag() for cls in union_members)],  # type: ignore # noqa: UP007
        TypedModel.discriminator(),
    ]



from __future__ import annotations

from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import (
    Generic,
    TypeVar,
)

from pvi._format.utils import Bounds
from pvi._format.widget import (
    GroupFormatter,
    SubScreenWidgetFormatter,
    WidgetFormatter,
    WidgetFormatterFactory,
    max_x,
    max_y,
    next_x,
    next_y,
)
from pvi.device import (
    ButtonPanel,
    Component,
    ComponentUnion,
    DeviceRef,
    Grid,
    Group,
    Row,
    SignalR,
    SignalRef,
    SignalRW,
    SignalW,
    SignalX,
    SubScreen,
    TableRead,
    TableWrite,
    Tree,
)

T = TypeVar("T")


@dataclass
class ScreenLayout:
    spacing: int
    title_height: int
    max_height: int
    group_label_height: int
    label_width: int
    widget_width: int
    widget_height: int
    group_widget_indent: int
    group_width_offset: int


@dataclass
class ScreenFormatterFactory(Generic[T]):
    screen_formatter_cls: type[GroupFormatter[T]]
    group_formatter_cls: type[GroupFormatter[T]]
    widget_formatter_factory: WidgetFormatterFactory[T]
    layout: ScreenLayout
    components: dict[str, ComponentUnion] = field(
        default_factory=dict[str, ComponentUnion]
    )
    base_file_name: str = ""

    def create_screen_formatter(
        self, components: Tree, title: str
    ) -> tuple[GroupFormatter[T], list[tuple[str, GroupFormatter[T]]]]:
        """Create an instance of `screen_cls` populated with widgets of `components`

        Args:
            components: A list of components that make up a device
            title: The title of the screen

        Returns:
            A `ScreenFormatter` plus a list of (file name, `ScreenFormatter`) for any
            nested `SubScreen`s in components

        """
        full_w = (
            self.layout.label_width + self.layout.widget_width + 2 * self.layout.spacing
        )
        screen_bounds = Bounds(h=self.layout.max_height)
        widget_dims = {"w": full_w, "h": self.layout.widget_height}
        screen_widgets: list[WidgetFormatter[T]] = []
        columns: list[Bounds] = [Bounds(**widget_dims)]

        match components:
            case [Group(layout=SubScreen()) as component]:
                # Expand single group sub screen as only component on screen
                # This is where the actual content of sub screens are created after
                # being replaced by a sub screen button in the original location
                components = component.children
            case _:
                pass

        for c in components:
            last_column_bounds = columns[-1]
            next_column_bounds = Bounds(
                x=next_x(screen_widgets, self.layout.spacing),
                y=0,
                **widget_dims,
            )
            if isinstance(c, Group) and not isinstance(c.layout, Row):
                # Create embedded group widget containing its components
                # Note: Group adjusts bounds to fit the components
                screen_widgets.extend(
                    self.create_group_formatters(
                        c,
                        screen_bounds=screen_bounds,
                        column_bounds=last_column_bounds,
                        next_column_bounds=next_column_bounds,
                    )
                )
            else:
                # Create top level single line widget or Row of widgets
                # Note: This will change columns in place
                screen_widgets.extend(
                    self.create_component_widget_formatters(
                        c,
                        parent_bounds=screen_bounds,
                        column_bounds=last_column_bounds,
                        next_column_bounds=next_column_bounds,
                        # Indent top-level widgets to align with Group widgets
                        indent=True,
                    )
                )

            if next_column_bounds.y != 0:
                columns.append(next_column_bounds)

        sub_screens = self.create_sub_screen_formatters(screen_widgets)

        screen_bounds.w = max_x(screen_widgets)
        screen_bounds.h = max_y(screen_widgets)
        return (
            self.screen_formatter_cls(
                bounds=screen_bounds, title=title, children=screen_widgets
            ),
            sub_screens,
        )

    def create_sub_screen_formatters(
        self, screen_widgets: list[WidgetFormatter[T]]
    ) -> list[tuple[str, GroupFormatter[T]]]:
        """Create and return `ScreenFormatter`s for any `SubScreenWidgetFormatters`

        When the root screen formatter is created it will format a button that opens a
        screen when a `SubScreen` widget is found, with a `SubScreenWidgetFormatter`.
        Here we search through the components of the screen for these formatters and
        create `ScreenFormatters`s that will format the screens that those buttons will
        open.

        This will recursively call `create_screen_formatter` and
        `create_sub_screen_formatters` until no further `SubScreenWidgetFormatters` are
        found.

        Args:
            screen_widgets: List of `WidgetFormatters` containing 0-or-more
            `SubScreenFactory`s to be found

        Returns:
            List of (file name, `ScreenFormatter`) created from `SubScreenFactory`s

        """
        sub_screen_widget_formatters = [
            # At the root
            widget_factory
            for widget_factory in screen_widgets
            if isinstance(widget_factory, SubScreenWidgetFormatter)
        ] + [
            # Nested in a Group
            group_widget_factory
            for widget_factory in screen_widgets
            if isinstance(widget_factory, GroupFormatter)
            for group_widget_factory in widget_factory.children
            if isinstance(group_widget_factory, SubScreenWidgetFormatter)
        ]

        sub_screen_formatters: list[tuple[str, GroupFormatter[T]]] = []
        for sub_screen_widget_formatter in sub_screen_widget_formatters:
            if sub_screen_widget_formatter.components is None:
                # This is a reference to an existing screen - don't create it
                continue

            factory: ScreenFormatterFactory[T] = ScreenFormatterFactory(
                screen_formatter_cls=self.screen_formatter_cls,
                group_formatter_cls=self.group_formatter_cls,
                widget_formatter_factory=self.widget_formatter_factory,
                layout=self.layout,
                base_file_name=f"{sub_screen_widget_formatter.file_name}",
            )
            screen_formatter, _sub_screen_formatters = factory.create_screen_formatter(
                [sub_screen_widget_formatter.components],
                sub_screen_widget_formatter.components.name,
            )
            sub_screen_formatters.append(
                (sub_screen_widget_formatter.file_name, screen_formatter)
            )
            sub_screen_formatters.extend(_sub_screen_formatters)

        return sub_screen_formatters

    def create_group_formatters(
        self,
        c: Group,
        screen_bounds: Bounds,
        column_bounds: Bounds,
        next_column_bounds: Bounds,
    ) -> list[WidgetFormatter[T]]:
        """Create widget formatters for a Group

        This could either be a list of widget formatters, or a single group formatter,
        depending on the layout.

        Args:
            c: Group of Component objects extracted from a device.yaml
            screen_bounds: Bounds of the top level screen
            column_bounds: Bounds of widget in current column
            next_column_bounds: Bounds of widget in next column should widgets exceed
                height limits

        Returns:
            Widget formatters representing the component

        Side Effect:
            One of {column_bounds,next_column_bounds}.y will be updated to fit the
            added widget

        """
        if is_table(c):
            # Make a sub screen button at the root to display this Group instead of
            # embedding the components within a Group widget
            c = Group(name=c.name, layout=SubScreen(), children=c.children)

        if isinstance(c.layout, SubScreen):
            return self.create_component_widget_formatters(
                Group(name=c.name, layout=SubScreen(), children=c.children),
                parent_bounds=screen_bounds,
                column_bounds=column_bounds,
                next_column_bounds=next_column_bounds,
                indent=True,
                add_label=False,
            )

        group_formatter = self.create_group_formatter(
            c, bounds=Bounds(x=column_bounds.x, y=column_bounds.y, h=screen_bounds.h)
        )

        if group_formatter.bounds.h + group_formatter.bounds.y <= screen_bounds.h:
            # Group fits in this column
            group_formatter.bounds.w += self.layout.group_width_offset

            # Update y for current column
            column_bounds.y = (
                group_formatter.bounds.y
                + group_formatter.bounds.h
                + self.layout.spacing
            )
        else:
            # Won't fit in first column, recreate in next column
            group_formatter = self.create_group_formatter(
                c,
                bounds=Bounds(
                    x=next_column_bounds.x, y=next_column_bounds.y, h=screen_bounds.h
                ),
            )
            group_formatter.bounds.w += self.layout.group_width_offset

            next_column_bounds.y = (
                group_formatter.bounds.y
                + group_formatter.bounds.h
                + self.layout.spacing
            )

        return [group_formatter]

    def create_group_formatter(self, group: Group, bounds: Bounds) -> GroupFormatter[T]:
        """Convert components within groups into widgets and lay them out within a
        group object.

        Args:
            group: Group of child Components in a Layout
            bounds: The size and position of the group Widget (x,y,w,h).

        Returns:
            A group object containing child widgets
        """

        full_w = (
            self.layout.label_width + self.layout.widget_width + 2 * self.layout.spacing
        )
        column_bounds = Bounds(w=full_w, h=self.layout.widget_height)
        widget_factories: list[WidgetFormatter[T]] = []

        assert isinstance(group.layout, Grid | SubScreen), (
            "Can only do Grid and SubScreen at the moment"
        )

        for c in group.children:
            component: Group | Component
            match c:
                case Group(layout=Grid()):
                    if is_table(c):
                        # Display table on a sub screen
                        component = Group(
                            name=c.name, layout=SubScreen(), children=c.children
                        )
                    else:
                        raise NotImplementedError(
                            "Cannot nest a Group(Grid()) in another Group on one screen"
                        )
                case _:
                    # Make single line with a component or Row of components
                    component = c

            next_column_bounds = Bounds(
                x=next_x(widget_factories, self.layout.spacing),
                w=full_w,
                h=self.layout.widget_height,
            )
            widget_factories.extend(
                self.create_component_widget_formatters(
                    component,
                    parent_bounds=bounds,
                    column_bounds=column_bounds,
                    next_column_bounds=next_column_bounds,
                    add_label=group.layout.labelled,
                )
            )
            if next_column_bounds.y != 0:
                # We have moved onto the next column
                column_bounds = next_column_bounds

        bounds.h = max_y(widget_factories)
        bounds.w = max_x(widget_factories)
        return self.group_formatter_cls(
            bounds=bounds, title=group.get_label(), children=widget_factories
        )

    def create_component_widget_formatters(
        self,
        c: ComponentUnion,
        parent_bounds: Bounds,
        column_bounds: Bounds,
        next_column_bounds: Bounds,
        indent: bool = False,
        add_label: bool = True,
    ) -> list[WidgetFormatter[T]]:
        """Generate widgets from component data and position them in a grid format

        Args:
            c: Component object extracted from a device.yaml
            column_bounds: Bounds of widget in current column
            parent_bounds: Size constraints from the object containing the widgets
            next_column_bounds: Bounds of widget in next column should widgets exceed
                height limits
            add_label: Whether the widget should have an associated label.
                Defaults to True.
            indent: Shift the resulting widgets to the group ident level
                Used for top-level widgets that are not inside a group.

        Returns:
            A collection of widgets representing the component

        Side Effect:
            One of {column_bounds,next_column_bounds}.y will be updated to fit the
            added widget

        """
        # Take copies so we don't modify the originals until we're done
        tmp_column_bounds = column_bounds.clone()
        tmp_next_column_bounds = next_column_bounds.clone()

        if indent:
            tmp_column_bounds.indent(self.layout.group_widget_indent)
            tmp_next_column_bounds.indent(self.layout.group_widget_indent)

        widgets = list(
            self.generate_component_formatters(c, tmp_column_bounds, add_label)
        )
        if max_y(widgets) <= parent_bounds.h:
            # Current column still fits on screen
            column_bounds.y = next_y(widgets, self.layout.spacing)
        else:
            # Widget makes current column too tall. Repeat in next column.
            widgets = list(
                self.generate_component_formatters(c, tmp_next_column_bounds, add_label)
            )
            next_column_bounds.y = next_y(widgets, self.layout.spacing)

        return widgets

    def generate_component_formatters(
        self,
        c: ComponentUnion,
        bounds: Bounds,
        add_label: bool = True,
    ) -> Iterator[WidgetFormatter[T]]:
        """Convert a component into its WidgetFormatter equivalents

        Args:
            c: Component object
            bounds: The size and position of the component widgets (x,y,w,h).
            add_label: Whether the component has an associated label. Defaults to True.

        Yields:
            A collection of widgets representing the component
        """

        # Widgets are allowed to expand bounds
        if not isinstance(c, SignalRef | Group):
            self.components[c.name] = c

        # Take a copy to modify in this scope
        component_bounds = bounds.clone()

        if isinstance(c, Group) and isinstance(c.layout, Row):
            # This Group should be formatted as a table
            if c.layout.header is not None:
                # Create table header
                assert len(c.layout.header) == len(c.children), (
                    "Header length does not match number of elements"
                )

                header_bounds = component_bounds.clone()
                if add_label:
                    # Scale the headers by the width with the label then
                    # indent them
                    original_width = len(c.layout.header) * (
                        header_bounds.w + self.layout.spacing
                    )
                    label_width = self.layout.label_width + self.layout.spacing

                    scaling_factor = (original_width - label_width) / original_width
                    header_bounds.w = int(header_bounds.w * scaling_factor)

                    header_bounds.indent(label_width)

                # Create column headers
                for column_header in c.layout.header:
                    yield self.widget_formatter_factory.header_formatter_cls(
                        bounds=header_bounds.clone(), text=column_header
                    )
                    header_bounds.x += header_bounds.w + self.layout.spacing

                # Shift down y
                component_bounds.y += self.layout.widget_height + self.layout.spacing

            row_components = c.children  # Create a widget for each row of Group
            # Allow given component width for each column, plus spacing
            component_bounds = component_bounds.tile(
                horizontal=len(c.children), spacing=self.layout.spacing
            )
        elif isinstance(c, SignalW) and isinstance(c.write_widget, ButtonPanel):
            # Convert SignalW into a SignalX with a fixed value for each action
            row_components = [
                SignalX(name=action, write_pv=c.write_pv, value=value)
                for action, value in c.write_widget.actions.items()
            ]
            # If this is a SignalRW, recreate the readback with a SignalR
            if isinstance(c, SignalRW) and c.read_widget is not None:
                row_components += [
                    SignalR(name=c.name, read_pv=c.read_pv, read_widget=c.read_widget)
                ]
        else:
            row_components = [c]  # Create one widget for row

            match c:
                case (
                    SignalR(read_widget=TableRead(widgets=widgets))
                    | SignalW(write_widget=TableWrite(widgets=widgets))
                ):
                    add_label = False  # Do not add row labels for Tables
                    component_bounds.w = 100 * len(widgets)
                    component_bounds.h *= 10  # TODO: How do we know the number of rows?
                case _:
                    pass

        if add_label:
            # Insert label and reduce width for widget
            left, row_bounds = component_bounds.split_left(
                self.layout.label_width, self.layout.spacing
            )
            yield self.widget_formatter_factory.label_formatter_cls(
                bounds=left,
                text=c.get_label(),
                tooltip=c.description or "No description provided",
            )
        else:
            # Allow full width for widget
            row_bounds = component_bounds

        if isinstance(c, SignalRef):
            yield from self.generate_component_formatters(
                self.components[c.name], row_bounds, add_label
            )
            return

        yield from self.generate_row_component_formatters(row_components, row_bounds)

    def generate_row_component_formatters(
        self,
        row_components: Sequence[Group | Component],
        row_bounds: Bounds,
    ) -> Iterator[WidgetFormatter[T]]:
        row_component_bounds = row_bounds.clone().split_into(
            len(row_components), self.layout.spacing
        )
        for rc_bounds, rc in zip(row_component_bounds, row_components, strict=True):
            # It is important to check for SignalX/SignalRW first, as they will also
            # match SignalR/SignalW

            if isinstance(rc, SignalX):
                yield self.widget_formatter_factory.action_formatter_cls(
                    bounds=rc_bounds,
                    label=rc.get_label(),
                    pv=rc.write_pv,
                    value=rc.value,
                )
            elif isinstance(rc, SignalRW):
                if rc.read_widget:
                    left, right = rc_bounds.split_into(2, self.layout.spacing)
                    yield from self.generate_write_widget(rc, left)
                    yield from self.generate_read_widget(rc, right)
                else:
                    yield from self.generate_write_widget(rc, rc_bounds)
            elif isinstance(rc, SignalW):
                yield from self.generate_write_widget(rc, rc_bounds)
            elif isinstance(rc, SignalR):
                yield from self.generate_read_widget(rc, rc_bounds)
            elif isinstance(rc, Group) and isinstance(rc.layout, SubScreen):
                yield self.widget_formatter_factory.sub_screen_formatter_cls(
                    bounds=rc_bounds,
                    label=rc.get_label(),
                    file_name=f"{self.base_file_name}_{rc.name.replace(' ', '_')}",
                    components=rc,
                )
            elif isinstance(rc, DeviceRef):
                yield self.widget_formatter_factory.sub_screen_formatter_cls(
                    bounds=rc_bounds,
                    label=" ".join(rc.macros.values()),
                    file_name=rc.ui,
                    macros=rc.macros,
                )
            else:
                print(f"Ignoring row component {rc}")

    def generate_read_widget(self, signal: SignalR, bounds: Bounds):
        if signal.read_widget is not None:
            yield self.widget_formatter_factory.pv_widget_formatter(
                signal.read_widget, bounds, signal.read_pv
            )

    def generate_write_widget(self, signal: SignalW, bounds: Bounds):
        yield self.widget_formatter_factory.pv_widget_formatter(
            signal.write_widget, bounds, signal.write_pv
        )


def is_table(component: Group) -> bool:
    return len(component.children) > 1 and all(
        isinstance(sub_component, Group) and isinstance(sub_component.layout, Row)
        for sub_component in component.children
    )
